import logging
import struct

logger = logging.getLogger("packetracer")

# Allows to activate/decative the auto-update of a header field.
# Note: _update_fields() has to be implemented
FIELD_FLAG_AUTOUPDATE = 1
# Identifies the header field defining the next higher layer type
# Allows to auto-set the value on concatenation: "pkt = eth0 + ip0" sets type field in eth0
# Auto-sets FIELD_FLAG_AUTOUPDATE
# Note: _update_fields() has to be implemented and _update_higherlayer_id() has to be called in it.
FIELD_FLAG_IS_TYPEFIELD = 2


def get_setter(varname, is_field_type_simple=True, is_field_static=True):
    """
    varname -- name of the variable to set the property for
    is_field_type_simple -- get property for simple static or dynamic type if True, else TriggerList
    is_field_static -- if is_field_type_simple is True: get static type (int, fixed size bytes, ...),
        else dynamic (format "xs") which can change in format (eg DNS names)

    return -- set-property for simple types or triggerlist
    """
    varname_shadowed = "_%s" % varname

    def setfield_simple(obj, value):
        """
        Set value for a simple field

        value -- bytes, int or None
        """
        if obj._unpacked is not None and not obj._unpacked:
            # obj._unpacked = None means: dissect not yet finished
            obj._unpack()
        if value is None and obj.__getattribute__(varname_shadowed + "_active"):
            # deactivate active field
            object.__setattr__(obj, varname_shadowed + "_active", False)
            obj._header_format_changed = True
        # logger.debug("deactivating field: %s" % varname_shadowed)
        elif value is not None and not obj.__getattribute__(varname_shadowed + "_active"):
            # activate inactive field
            object.__setattr__(obj, varname_shadowed + "_active", True)
            obj._header_format_changed = True
        # logger.debug("activating field: %s" % varname_shadowed)

        if value is not None and not is_field_static:
            # simple dynamic field: update format
            format_new = "%ds" % len(value)
            # logger.debug(">>> changing format for dynamic field: %r / %s / %s" %
            # (obj.__class__, varname_shadowed, format_new))
            object.__setattr__(obj, varname_shadowed + "_format", format_new)
            obj._header_format_changed = True

        # logger.debug("setting simple field: %r=%r" % (varname_shadowed, value))
        object.__setattr__(obj, varname_shadowed, value)
        obj._header_changed = True
        obj._notify_changelistener()

    def setfield_triggerlist(obj, value):
        """
        Clear list and add value as only value.

        value -- Packet, bytes (single or as list)
        """
        tl = obj.__getattribute__(varname_shadowed)

        if type(tl) is list:
            # We need to create the original TriggerList in order to unpack correctly
            # _triggerlistName = [b"bytes", callback] or
            # _triggerlistName = [b"", callback] (default initiation)
            # logger.debug(">>> initiating TriggerList")
            tl = obj._header_fields_dyn_dict[varname_shadowed](
                obj,
                dissect_callback=tl[1],
                buffer=tl[0],
                headerfield_name=varname_shadowed
            )
            object.__setattr__(obj, varname_shadowed, tl)
        # This will trigger unpacking

        del tl[:]

        # TriggerList: avoid overwriting dynamic fields eg when using keyword constructor Class(key=val)
        if type(value) is list:
            tl.extend(value)
        else:
            tl.append(value)
        obj._header_changed = True
        obj._notify_changelistener()

    if is_field_type_simple:
        return setfield_simple

    return setfield_triggerlist


def get_getter(varname, is_field_type_simple=True):
    """
    varname -- name of the variable to set the property for
    is_field_type_simple -- get property for simple static or dynamic type if True, else TriggerList
    return -- get-property for simple type or triggerlist
    """
    varname_shadowed = "_%s" % varname

    def getfield_simple(obj):
        """
        Unpack field ondemand
        """
        # logger.debug("getting value for simple field: %s" % varname_shadowed)
        if obj._unpacked is not None and not obj._unpacked:
            obj._unpack()
        # logger.debug("getting simple field: %r=%r" %
        # (varname_shadowed, obj.__getattribute__(varname_shadowed)))
        return obj.__getattribute__(varname_shadowed)

    def getfield_triggerlist(obj):
        tl = obj.__getattribute__(varname_shadowed)
        # logger.debug(">>> getting Triggerlist for %r: %r" % (obj.__class__, tl))

        if type(tl) is list:
            # _triggerlistName = [b"bytes", callback] or
            # _triggerlistName = [b"", callback] (default initiation)
            tl = obj._header_fields_dyn_dict[varname_shadowed](
                obj,
                dissect_callback=tl[1],
                buffer=tl[0],
                headerfield_name=varname_shadowed
            )
            object.__setattr__(obj, varname_shadowed, tl)

        return tl

    if is_field_type_simple:
        return getfield_simple

    return getfield_triggerlist


def configure_packet_header(t, hdrs, header_fmt):
    if hdrs is None:
        return

    # Create a property for every field: property a -> get/set access to _a_shadowed.
    # Using properties will slow down access to header fields but it's needed:
    # This way we get informed about get-access (needed to check for unpack)
    # more efficiently than using __getattribute__ (slow access for header
    # fields vs. slow access for ALL class members).
    # Every header field will get two additional values set:
    # var_active = indicates if header is active
    # var_format = indicates the header format
    for hdr in hdrs:
        # Sanity checks
        # Max packet fields: [name, format, value, flags, {value:key})
        if len(hdr) > 4:
            logger.warning("Amount of field definitions > 4: %r", hdr)

        if hdr[1] is not None:
            try:
                struct.Struct(hdr[1])
            except struct.error:
                raise Exception("Invalid format specified in class %s for header '%s': '%s'" %
                                (t.__module__ + "." + t.__name__, hdr[0], hdr[1]))

            if hdr[2] is not None:
                try:
                    struct.Struct(hdr[1]).pack(hdr[2])
                except struct.error:
                    raise Exception("Invalid value specified in class %s for header '%s' with format '%s': '%s'" %
                                    (t.__module__ + "." + t.__name__, hdr[0], hdr[1], hdr[2]))

        shadowed_name = "_%s" % hdr[0]
        t._header_field_names.append(shadowed_name)
        setattr(t, shadowed_name + "_active", True)

        # remember header format
        # t._header_field_infos[shadowed_name] = [True, hdr[1]]
        is_field_type_simple = False
        is_field_static = True

        if hdr[1] is not None or (hdr[2] is None or type(hdr[2]) == bytes):
            # simple static or simple dynamic type
            # we got one of: ("name", format, ???) = static or
            # ("name", None, [None, b"xxx"]) = dynamic
            # -> Format given = static, Format None = dynamic
            is_field_type_simple = True

            if hdr[1] is None:
                # assume simple dynamic field
                is_field_static = False

        setattr(t, shadowed_name + "_format", hdr[1])

        if is_field_type_simple:
            # assume simple static or simple dynamic type
            fmt = hdr[1]

            if hdr[2] is not None:
                # value given: field is active
                if fmt is None:
                    # dynamic field
                    fmt = "%ds" % len(hdr[2])
                    setattr(t, shadowed_name + "_format", fmt)
                header_fmt.append(fmt)
                t._header_cached.append(hdr[2])
            # logger.debug("--------> field is active: %r" % hdr[0])
            # else:
            #	setattr(t, shadowed_name + "_active", False)

            # only simple fields can get deactivated
            setattr(t, shadowed_name + "_active", True if hdr[2] is not None else False)

            # check for auto-update
            if len(hdr) >= 4:
                field_flags = hdr[3]

                if field_flags & FIELD_FLAG_IS_TYPEFIELD != 0:
                    # logger.debug("setting _id_fieldname: %r" % (hdr[0]))
                    setattr(t, "_id_fieldname", hdr[0])
                    # xxx__au_active must be set: read by _update_higherlayer_id
                    field_flags |= FIELD_FLAG_AUTOUPDATE

                if field_flags & FIELD_FLAG_AUTOUPDATE != 0:
                    # logger.debug("marking %s as auto-update" % hdr[0])
                    # remember which fields are auto-update ones, default is active
                    setattr(t, hdr[0] + "_au_active", True)

            # Setting/getting value is done via properties.
            # Set initial value via shadowed variable:
            # _varname <- varname [optional in subclass: <- varname_s]
            # logger.debug("init simple type: %s=%r" % (shadowed_name, hdr[2]))
            setattr(t, shadowed_name, hdr[2])
            setattr(t, hdr[0], property(
                get_getter(hdr[0], is_field_type_simple=True),
                get_setter(hdr[0], is_field_type_simple=True, is_field_static=is_field_static)
            ))
        else:
            # assume TriggerList
            # Triggerlists don't have initial default values (and can't get deactivated)
            t._header_fields_dyn_dict[shadowed_name] = hdr[2]
            # initial value of TiggerLists is: values to init empty list
            setattr(t, shadowed_name, [b"", None])
            setattr(t, hdr[0], property(
                get_getter(hdr[0], is_field_type_simple=False),
                get_setter(hdr[0], is_field_type_simple=False, is_field_static=is_field_static)
            ))
            # format and value needed for correct length in _unpack()
            # default is empty TriggerList -> 0 bytes
            header_fmt.append("0s")
            t._header_cached.append(b"")


def configure_packet_header_sub(t, hdrs_sub):
    if hdrs_sub is None:
        return

    for name_cbget_cbset in hdrs_sub:
        if len(name_cbget_cbset) < 2:
            logger.warning("subheader length < 2: %d", len(name_cbget_cbset))
            continue
        # logger.debug("setting subheader: %s", name_cbget_cbset[0])

        # (name, cb_get, cb_set)
        if len(name_cbget_cbset) == 3:
            setattr(t, name_cbget_cbset[0], property(name_cbget_cbset[1], name_cbget_cbset[2]))
        # (name, cb_get)
        else:
            setattr(t, name_cbget_cbset[0], property(name_cbget_cbset[1]))


class MetaPacket(type):
    """
    This Metaclass is a more efficient way of setting attributes than using __init__.
    This is done by reading name, format and default value out of a mendatory __hdr__
    tuple in every subclass. This configuration is set one time when loading the module
    (not at instantiation). Attributes can be normally accessed using "obj.field" notation.
    General note: Callflaw is: __new__ (loading module) -> __init__ (initiate class)

    Header defintition example:
    __hdr__ = (
        ("header1", "H", 123), # simple static field
        ("header2", "H", None), # simple static field, inactive
        ("header3", None, b"xxx"), # simple dynamic field
        ("header4", None, None), # simple dynamic field, inactive
        ("header5", None, Triggerlist) # TriggerList field
    )

    For values <1 byte a subheader definition eases up setting/getting those values:

    __hdr_sub__ = (
        ("header1_sub",
            lambda val: val & 1							# callback to retrieve value
            lambda obj, val: obj.__setattr__(val & 1)	# callback to set value
        ),
        ...
    )

    CAUTION:
    - List et al are _SHARED_ among all instantiated classes! A copy is needed on
    changes to them without side effects
    - New protocols: header field names must be unique among other variable and method names
    """

    def __new__(cls, clsname, clsbases, clsdict):
        # Slots (dct["__slots__"] = ...) can't be used because:
        # Setting default values (eg for _header_fields_dyn_dict) must
        # be done in __init__ which increases delay (init for every instantiation...)
        # Sidenote: Setting default values here creates readonly exception later:
        # __slots__ = ("var", ...) -> t.var = None -> p = Clz() -> p.var = 123 won't work (var is readonly)
        # See: https://stackoverflow.com/questions/820671/python-slots-and-attribute-is-read-only
        t = type.__new__(cls, clsname, clsbases, clsdict)
        # dictionary of TriggerLists: name -> TriggerListClass
        t._header_fields_dyn_dict = {}
        # cache header for performance reasons, will be set to bytes later on
        t._header_cached = []
        # all header names
        t._header_field_names = []
        # all header formats including byte order
        header_fmt = [">"]

        # varname holding the fieldname containing the id associated with body handler
        # eg Ethernet -> "type" or IP -> "p"
        t._id_fieldname = None

        # get header-infos: [("name", "format", value), ...]
        hdrs = getattr(t, "__hdr__", None)
        configure_packet_header(t, hdrs, header_fmt)

        # get sub-header-infos: [("name", cb_get, cb_set), ...]
        hdrs_sub = getattr(t, "__hdr_sub__", None)
        configure_packet_header_sub(t, hdrs_sub)

        # get handler classes, assume Packet class has no member "__handler__"
        handler = getattr(t, "__handler__", None)

        if handler is not None and len(handler) > 0:
            if handler.__class__ is not dict:
                print("Invalid format of __handler__: not a dictionary! %r", handler)
            else:
                t.load_handler(t, handler)

        # Set of higher_layer classes which force dissecting (update dependends):
        # IP->TCP => IP changed, TCP needs checksum update and
        # needs to be dissected for this => TCP is ud in ip
        update_deps = getattr(t, "UPDATE_DEPENDANTS", set())
        t._update_dependants = set([ud.__class__ for ud in update_deps])

        # logger.debug(">>> translated header names: %s/%r" % (clsname, t._header_name_translate))
        # current format as string
        t._header_format = struct.Struct("".join(header_fmt))
        # header size can be assigened by __init__() directly or given by _header_format.size
        t._header_len = t._header_format.size
        # track changes to header format (changes to simple dynamic fields or TriggerList)
        t._header_format_changed = False
        # cached header, return this if nothing changed
        t._header_cached = t._header_format.pack(*t._header_cached)
        # logger.debug("formatstring is: %s" % header_fmt)
        # body as raw byte string (None if handler is present)
        t._body_bytes = b""
        # next lower layer: a = b + c -> b will be lower layer for c
        t._lower_layer = None
        t._higher_layer = None
        # track changes to header values: This is needed for layers like TCP for
        # checksum-recalculation. Set to "True" on changes to header/body values, set to False on "bin()"
        # track changes to header values
        t._header_changed = False
        # track changes to body value like [None | bytes | body-handler] -> [None | bytes | body-handler]
        t._body_changed = False
        # objects which get notified on changes on header or body (shared)
        # needs to be None do identify none-initialized variable
        t._changelistener = None
        # parent of the packet which is contained in a triggerlist
        # parent_packet.triggerlist[sub] -> sub._triggelistpacket_parent == parent_packet
        t._triggelistpacket_parent = None
        # lazy handler data: [name, class, bytes]
        t._lazy_handler_data = None
        # Indicates the most top layer until which should be unpacked
        # (vs. lazy dissecting = just next higher layer)
        # Setting this to an unknown class will keep the next-layer-parsing going on
        t._final_unpack_clz = None
        # indicates if static header values got already unpacked
        # [True|False] = Status after dissect, None = pre-dissect (not unpacked)
        t._unpacked = None
        # indicates if this packet contains fragmented data saved as body bytes
        t._fragmented = False
        # concatination of errors, see packetracer.py -> ERROR_...
        t._errors = 0
        return t



"""Dynamic Host Configuration Protocol."""
import logging

from packetracer import packetracer, triggerlist
from packetracer.layer12 import arp

logger = logging.getLogger("packetracer")

DHCP_OP_REQUEST = 1
DHCP_OP_REPLY = 2

DHCP_MAGIC = 0x63825363

# DHCP option codes
DHCP_OPT_NETMASK = 1  # I: subnet mask
DHCP_OPT_TIMEOFFSET = 2
DHCP_OPT_ROUTER = 3  # s: list of router ips
DHCP_OPT_TIMESERVER = 4
DHCP_OPT_NAMESERVER = 5
DHCP_OPT_DNS_SVRS = 6  # s: list of DNS servers
DHCP_OPT_LOGSERV = 7
DHCP_OPT_COOKIESERV = 8
DHCP_OPT_LPRSERV = 9
DHCP_OPT_IMPSERV = 10
DHCP_OPT_RESSERV = 11
DHCP_OPT_HOSTNAME = 12  # s: client hostname
DHCP_OPT_BOOTFILESIZE = 13
DHCP_OPT_DUMPFILE = 14
DHCP_OPT_DOMAIN = 15  # s: domain name
DHCP_OPT_SWAPSERV = 16
DHCP_OPT_ROOTPATH = 17
DHCP_OPT_EXTENPATH = 18
DHCP_OPT_IPFORWARD = 19
DHCP_OPT_SRCROUTE = 20
DHCP_OPT_POLICYFILTER = 21
DHCP_OPT_MAXASMSIZE = 22
DHCP_OPT_IPTTL = 23
DHCP_OPT_MTUTIMEOUT = 24
DHCP_OPT_MTUTABLE = 25
DHCP_OPT_MTUSIZE = 26
DHCP_OPT_LOCALSUBNETS = 27
DHCP_OPT_BROADCASTADDR = 28
DHCP_OPT_DOMASKDISCOV = 29
DHCP_OPT_MASKSUPPLY = 30
DHCP_OPT_DOROUTEDISC = 31
DHCP_OPT_ROUTERSOLICIT = 32
DHCP_OPT_STATICROUTE = 33
DHCP_OPT_TRAILERENCAP = 34
DHCP_OPT_ARPTIMEOUT = 35
DHCP_OPT_ETHERENCAP = 36
DHCP_OPT_TCPTTL = 37
DHCP_OPT_TCPKEEPALIVE = 38
DHCP_OPT_TCPALIVEGARBAGE = 39
DHCP_OPT_NISDOMAIN = 40
DHCP_OPT_NISSERVERS = 41
DHCP_OPT_NISTIMESERV = 42
DHCP_OPT_VENDSPECIFIC = 43
DHCP_OPT_NBNS = 44
DHCP_OPT_NBDD = 45
DHCP_OPT_NBTCPIP = 46
DHCP_OPT_NBTCPSCOPE = 47
DHCP_OPT_XFONT = 48
DHCP_OPT_XDISPLAYMGR = 49
DHCP_OPT_REQ_IP = 50  # I: IP address
DHCP_OPT_LEASE_SEC = 51  # I: lease seconds
DHCP_OPT_OPTIONOVERLOAD = 52
DHCP_OPT_MSGTYPE = 53  # B: message type
DHCP_OPT_SERVER_ID = 54  # I: server IP address
DHCP_OPT_PARAM_REQ = 55  # s: list of option codes
DHCP_OPT_MESSAGE = 56
DHCP_OPT_MAXMSGSIZE = 57
DHCP_OPT_RENEWTIME = 58
DHCP_OPT_REBINDTIME = 59
DHCP_OPT_VENDOR_ID = 60  # s: vendor class id
DHCP_OPT_CLIENT_ID = 61  # Bs: idtype, id (idtype 0: FQDN, idtype 1: M
DHCP_OPT_NISPLUSDOMAIN = 64
DHCP_OPT_NISPLUSSERVERS = 65
DHCP_OPT_MOBILEIPAGENT = 68
DHCP_OPT_SMTPSERVER = 69
DHCP_OPT_POP3SERVER = 70
DHCP_OPT_NNTPSERVER = 71
DHCP_OPT_WWWSERVER = 72
DHCP_OPT_FINGERSERVER = 73
DHCP_OPT_IRCSERVER = 74
DHCP_OPT_STSERVER = 75
DHCP_OPT_STDASERVER = 76

# DHCP message type values
DHCPDISCOVER = 1
DHCPOFFER = 2
DHCPREQUEST = 3
DHCPDECLINE = 4
DHCPACK = 5
DHCPNAK = 6
DHCPRELEASE = 7
DHCPINFORM = 8


class DHCP(packetracer.Packet):
    __hdr__ = (
        ("op", "B", DHCP_OP_REQUEST),
        ("hrd", "B", arp.ARP_HRD_ETH),  # just like ARP.hrd
        ("hln", "B", 6),  # and ARP.hln
        ("hops", "B", 0),
        ("xid", "I", 0xDEADBEEF),
        ("secs", "H", 0),
        ("flags", "H", 0),
        ("ciaddr", "4s", b"\x00" * 4),
        ("yiaddr", "4s", b"\x00" * 4),
        ("siaddr", "4s", b"\x00" * 4),
        ("giaddr", "4s", b"\x00" * 4),
        # MAC + padding
        ("chaddr", "16s", b"\x00" * 6 + b"\x00" * 10),
        ("sname", "64s", b"\x00" * 64),
        ("file", "128s", b"\x00" * 128),
        ("magic", "I", DHCP_MAGIC),
        ("opts", None, triggerlist.TriggerList)
    )

    ciaddr_s = packetracer.get_property_ip4("ciaddr")
    yiaddr_s = packetracer.get_property_ip4("yiaddr")
    siaddr_s = packetracer.get_property_ip4("siaddr")
    giaddr_s = packetracer.get_property_ip4("giaddr")

    def _dissect(self, buf):
        # logger.debug("DHCP: parsing options, buflen: %d" % len(buf))
        self._init_triggerlist("opts", buf[28 + 16 + 64 + 128 + 4:], DHCP.__get_opts)
        # logger.debug(buf[28+16+64+128+4:])
        # logger.debug("amount of options after parsing: %d" % len(self.opts))
        return len(buf)

    @staticmethod
    def __get_opts(buf):
        # logger.debug("DHCP: parsing options from: %s" % buf)
        opts = []
        i = 0

        while i < len(buf):
            t = buf[i]
            p = None
            # logger.debug("DHCP: adding option type %d" % t)

            # last option
            if t in [0, 0xFF]:
                p = DHCPOpt(type=t, len=0)
                i += 1
            else:
                dlen = buf[i + 1]
                p = DHCPOpt(type=t, len=dlen, body_bytes=buf[i + 2: i + 2 + dlen])
                i += 2 + dlen

            # logger.debug("new option: %s" % p)
            opts.append(p)

            if t == 0xFF:
                if i < len(buf):
                    # padding is part of the options
                    opts.append(Padding(buf[i:]))
                break

        return opts


class DHCPOpt(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
        ("len", "B", 0),
    )


class Padding(packetracer.Packet):
    pass



"""
ISO15118
"""
import logging

from packetracer.packetracer import Packet

logger = logging.getLogger("packetracer")

MSGTYPE_EXI = 0x8001
MSGTYPE_SDP_REQ = 0x9000
MSGTYPE_SDP_RSP = 0x9001


class SDP(Packet):
    __hdr__ = (
        ("id", "B", 0),
        ("idrev", "B", 0),
        ("msgtype", "H", 0),
        ("msglen", "I", 0)
    )



"""
Scalable service-Oriented MiddlewarE over IP (SOME/IP)
https://www.autosar.org/fileadmin/user_upload/standards/foundation/1-2/AUTOSAR_PRS_SOMEIPServiceDiscoveryProtocol.pdf
"""
import logging

from packetracer.packetracer import Packet
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE

logger = logging.getLogger("packetracer")

RET_CODE_E_OK = 0x00
RET_CODE_E_NOT_OK = 0x01
RET_CODE_E_UNKNOWN_SERVICE = 0x02
RET_CODE_E_UNKNOWN_METHOD = 0x03
RET_CODE_E_NOT_READY = 0x04
RET_CODE_E_NOT_REACHABLE = 0x05
RET_CODE_E_TIMEOUT = 0x06
RET_CODE_E_WRONG_PROTOCOL_VERSION = 0x07
RET_CODE_E_WRONG_INTERFACE_VERSION = 0x08
RET_CODE_E_MALFORMED_MESSAGE = 0x09
RET_CODE_E_WRONG_MESSAGE_TYPE = 0x0A


# Other codes: RESERVED


class SomeIP(Packet):
    __hdr__ = (
        ("messageid", "I", 1),
        ("length", "I", 8, FIELD_FLAG_AUTOUPDATE),  # in bytes, inclusive 8 bytes of header
        ("reqid", "I", 0),
        ("protoversion", "B", 0),
        ("ifaceversion", "B", 0),
        ("msgtype", "B", 0),
        ("retcode", "B", 0)
    )

    def _update_fields(self):
        if not self._changed():
            return

        if self.length_au_active:
            self.length = 8 + len(self.body_bytes)



"""
Signal Level Attenuation Characterization (SLAC)
HomePlug Green PHY Specification
"""
import logging
import sys

from packetracer import packetracer
from packetracer.packetracer import Packet
from packetracer.structcbs import pack_H_le, unpack_H, unpack_H_le
from packetracer.triggerlist import TriggerList

logger = logging.getLogger("packetracer")

MASK_MSGTYPE_LE = 0xFCF9
MASK_MMTYPELSB_LE = 0x0300
MASK_MMTYPEMSB_LE = 0x0006

module_this = sys.modules[__name__]

# who defined all that useless messages???
TYPEINFO_DESCRIPTION = {
    # Central <-> Station
    0x0000: "CC_CCO_APPOINT",
    0x0004: "CC_BACKUP_APP",
    0x0008: "CC_LINK_INFO",
    0x000C: "CC_HANDOVER",
    0x0010: "CC_HANDOVER_INFO",
    0x0014: "CC_DISCOVER_LIST",
    0x0018: "CC_LINK_NEW",
    0x001C: "CC_LINK_MOD",
    0x0020: "CC_LINK_SQZ",
    0x0024: "CC_LINK_REL",
    0x0028: "CC_DETECT_REPORT",
    0x002C: "CC_WHO_RU",
    0x0030: "CC_ASSOC",
    0x0034: "CC_LEAVE",
    0x0038: "CC_SET_TEI_MAP",
    0x003C: "CC_RELAY",
    0x0040: "CC_BEACON_RELIABILITY.REQ",
    0x0044: "CC_ALLOC_MOVE",
    0x0048: "CC_ACCESS_NEW",
    0x004C: "CC_ACCESS_REL",
    0x0050: "CC_DCPPC",
    0x0054: "CC_HP1_DET",
    0x0058: "CC_BLE_UPDATE",
    0x005C: "CC_BCAST_REPEAT",
    0x0060: "CC_MH_LINK_NEW",
    0x0064: "CC_ISP_DetectionReport.IND",
    0x0068: "CC_ISP_StartReSync",
    0x006C: "CC_ISP_FinishReSync",
    0x0070: "CC_ISP_ReSyncDetected.IND",
    0x0074: "CC_ISP_ReSyncTransmit.REQ",
    0x0078: "CC_POWERSAVE.",
    0x007C: "CC_POWERSAVE_EXIT.REQ",
    0x0080: "CC_POWERSAVE_LIST.REQ",
    0x0084: "CC_STOP_POW",
    # Proxy Coordinator
    0x2000: "CP_PROXY_APPOINT",
    0x2004: "PH_PROXY_APPOINT",
    0x2008: "CP_PROXY_WAKE.",
    # CCo - CCo
    0x4000: "NN_INL.REQ",
    0x4004: "NN_NEW_NET.RE",
    0x4008: "NN_ADD_ALLOC.R",
    0x400C: "NN_REL_ALLOC.R",
    0x4010: "NN_REL_NET.IND",
    # Station - Station
    0x6000: "CM_UNASSOCIATED",
    0x6004: "CM_ENCRYPTED_PAYLOAD",
    0x6008: "CM_SET_KEY",
    0x600C: "CM_GET_KEY",
    0x6010: "CM_SC_JOIN",
    0x6014: "CM_CHAN_EST",
    0x6018: "CM_TM_UPDATE",
    0x601C: "CM_AMP_MAP",
    0x6020: "CM_BRG_INFO",
    0x6024: "CM_CONN_NEW",
    0x6028: "CM_CONN_REL",
    0x602C: "CM_CONN_MOD",
    0x6030: "CM_CONN_INFO",
    0x6034: "CM_STA_CAP",
    0x6038: "CM_NW_INFO",
    0x603C: "CM_GET_BEACON",
    0x6040: "CM_HFID",
    0x6044: "CM_MME_ERROR",
    0x6048: "CM_NW_STATS",
    0x604C: "CM_LINK_STATS",
    0x6050: "CM_ROUTE_INFO",
    0x6054: "CM_UNREACHABLE",
    0x6058: "CM_MH_CONN_NEW",
    0x605C: "CM_EXTENDED_TONEMASK",
    0x6060: "CM_STA_IDENTIFY",
    0x6064: "CM_SLAC_PARM",
    0x6068: "CM_START_ATTEN_CHAR",
    0x606C: "CM_ATTEN_CHAR",
    0x6070: "CM_PKCS_CERT",
    0x6074: "CM_MNBC_SOUND",
    0x6078: "CM_VALIDATE",
    0x607C: "CM_SLAC_MATCH",
    0x6080: "CM_SLAC_USER_DATA",
    0x6084: "CM_ATTEN_PROFILE",
    0xA0B8: "VS_PL_LNK_STATUS"
}

# reverse access of message IDs
for msgid, name in TYPEINFO_DESCRIPTION.items():
    setattr(module_this, name, msgid)

# Management message type LSB
MMTYPE_LSB_DESCRIPTION = {
    0x00: "MMTYPELSB_REQUEST",
    0x01: "MMTYPELSB_CONFIRM",
    0x02: "MMTYPELSB_INDICATION",
    0x03: "MMTYPELSB_RESPONSE"
}

for msgid, name in MMTYPE_LSB_DESCRIPTION.items():
    setattr(module_this, name, msgid)

# Management message type MSB
MMTYPE_MSB_DESCRIPTION = {
    0x00: "MMTYPEMSB_STA__CentralCoordinator",
    0x01: "MMTYPEMSB_ProxyCoordinator",
    0x02: "MMTYPEMSB_CentralCoordinator__CentralCoordinator",
    0x03: "MMTYPEMSB_STA__STA",
    0x04: "MMTYPEMSB_Manufacturer_Specific"
}

for msgid, name in MMTYPE_MSB_DESCRIPTION.items():
    setattr(module_this, name, msgid)


class CMSetKeyReq(Packet):
    __hdr__ = (
        ("keytype", "B", 0),
        ("mynonce", "I", 0),
        ("yournonce", "I", 0),
        ("pid", "B", 0),
        ("prn", "H", 0),
        ("pwm", "B", 0),
        ("ccocapa", "B", 0),
        ("nid", "7s", b"\x00" * 7),
        ("neweks", "B", 0),
        # length = 0, 16
        ("newkey", None, None),
    )


class CMSetKeyICnf(Packet):
    __hdr__ = (
        ("result", "B", 0),
        ("mynonce", "I", 0),
        ("yournonce", "I", 0),
        ("pid", "B", 0),
        ("prn", "H", 0),
        ("pwm", "B", 0),
        ("ccocapa", "B", 0)
    )


class CMAttenCharInd(Packet):
    __hdr__ = (
        ("apptype", "B", 0),
        ("sectype", "B", 0),
        ("sourceaddr", "6s", b"\x00" * 6),
        ("runid", "Q", 0),
        ("sourceid", "17s", b"\x00" * 17),
        ("respid", "17s", b"\x00" * 17),
        ("numsounds", "B", 0)
    )

    sourceaddr_s = packetracer.get_property_mac("sourceaddr")


class CMAttenCharRsp(Packet):
    __hdr__ = (
        ("apptype", "B", 0),
        ("sectype", "B", 0),
        ("sourceaddr", "6s", b"\x00" * 6),
        ("runid", "Q", 0),
        ("sourceid", "17s", b"\x00" * 17),
        ("respid", "17s", b"\x00" * 17),
        ("result", "B", 0)
    )

    sourceaddr_s = packetracer.get_property_mac("sourceaddr")


class CMSlacParmReq(Packet):
    __hdr__ = (
        ("apptype", "B", 0),
        ("sectype", "B", 0),
        ("runid", "Q", 0),
        # Only present if security type is 1
        ("ciphersuitesize", "B", None),
        ("ciphersuites", None, TriggerList)
    )


class CMSlacParmCnf(Packet):
    __hdr__ = (
        ("msoundtarget", "6s", b"\x00" * 6),
        ("numsounds", "B", 0),
        ("timeout", "B", 0),
        ("resptype", "B", 0),
        ("forwardingsta", "6s", b"\x00" * 6),
        ("apptype", "B", 0),
        ("sectype", "B", 0),
        ("runid", "Q", 0),
        # Only present if security type is 1
        ("ciphersuite", "H", None)
    )

    msoundtarget_s = packetracer.get_property_mac("msoundtarget")
    forwardingsta_s = packetracer.get_property_mac("forwardingsta")


class CMStartAttenCharInd(Packet):
    __hdr__ = (
        ("apptype", "B", 0),
        ("sectype", "B", 0),
        ("numsounds", "B", 0),
        ("timeout", "B", 0),
        ("resptype", "B", 0),
        ("forwardingsta", "6s", b"\x00" * 6),
        ("runid", "Q", 0),
    )

    forwardingsta_s = packetracer.get_property_mac("forwardingsta")


class CMMnbcSoundInd(Packet):
    __hdr__ = (
        ("apptype", "B", 0),
        ("sectype", "B", 0),
        ("senderid", "17s", b"\x00" * 17),
        ("cnt", "B", 0),
        ("runid", "Q", 0),
        ("rsvd", "8s", b"\x00" * 8),
        ("rnd", "16s", b"\x00" * 16)
    )


class CMSlacMatchReq(Packet):
    __hdr__ = (
        ("apptype", "B", 0),
        ("sectype", "B", 0),
        ("mvflen", "H", 0),
        ("pevid", "17s", b"\x00" * 17),
        ("pevmac", "6s", b"\x00" * 6),
        ("evseid", "17s", b"\x00" * 17),
        ("evsemac", "6s", b"\x00" * 6),
        ("runid", "Q", 0),
        ("rsvd", "8s", b"\x00" * 8)
    )

    def _get_mvflen_be(self):
        return unpack_H(pack_H_le(self.mvflen))[0]

    def _set_mvflen_be(self, val):
        self.mvflen = unpack_H(pack_H_le(val))[0]

    mvflen_be = property(_get_mvflen_be, _set_mvflen_be)

    pevmac_s = packetracer.get_property_mac("pevmac")
    evsemac_s = packetracer.get_property_mac("evsemac")


class CMSlacMatchCnf(Packet):
    __hdr__ = (
        ("apptype", "B", 0),
        ("sectype", "B", 0),
        ("mvflen", "H", 0),
        ("pevid", "17s", b"\x00" * 17),
        ("pevmac", "6s", b"\x00" * 6),
        ("evseid", "17s", b"\x00" * 17),
        ("evsemac", "6s", b"\x00" * 6),
        ("runid", "Q", 0),
        ("rsvd1", "8s", b"\x00" * 8),
        ("nid", "7s", b"\x00" * 7),
        ("rsvd2", "B", 0),
        ("nmk", "16s", b"\x00" * 16),
    )

    pevmac_s = packetracer.get_property_mac("pevmac")
    evsemac_s = packetracer.get_property_mac("evsemac")


class CMLinkStatsReq(Packet):
    __hdr__ = (
        ("reqtype", "B", 0),
        ("reqid", "B", 0),
        ("nid", "7s", b"\x00" * 7),
        ("lid", "B", 0),
        ("tlflag", "B", 0),
        ("mgmtflag", "B", 0),
        ("dasa", "6s", b"\x00" * 6)
    )

    dasa_s = packetracer.get_property_mac("dasa")


# TODO: LinkStats payload as handler for CMLinkStatsCnf, how to differentiate?

class CMLinkStatsCnf(Packet):
    __hdr__ = (
        ("reqid", "B", 0),
        ("restype", "B", 0),
        ("linkstats", "H", 0)
    )


class VSPLLinkStatusReq(Packet):
    __hdr__ = (
        ("oui", "3s", b"\x00" * 3),
    )


class VSPLLinkStatusCnf(Packet):
    __hdr__ = (
        ("oui", "3s", b"\x00" * 3),
        ("link", "H", 0),
    )


class AMPMapReq(Packet):
    __hdr__ = (
        ("amlen", "H", 0),
    )


class AMPMapCnf(Packet):
    __hdr__ = (
        ("restype", "B", 0),
    )


class CMPKCSCertReq(Packet):
    __hdr__ = (
        ("targetmac", "6s", b"\x00" * 6),
        ("ciphersuitesize", "B", 0),
        ("cipersuite", None, TriggerList)
    )

    targetmac_s = packetracer.get_property_mac("targetmac")


class CMPKCSCertCnf(Packet):
    __hdr__ = (
        ("targetmac", "6s", b"\x00" * 6),
        ("status", "B", 0),
        ("cipersuite", "H", 0),
        ("certlen", "H", 0),
        ("certpackage", None, TriggerList)
    )

    targetmac_s = packetracer.get_property_mac("targetmac")


class CMPKCSCertInd(Packet):
    """
    When the CM_SLAC_PARM.CNF indicates that Secure SLAC is required, the PEV-HLE
    shall send a CM_PKCS_CERT.IND message. The Target MAC address for this message
    shall be set to MAC address of the PEV Green PHY station. To ensure reliable
    reception of this message at all EVSEs, it is recommended that this message be
    transmitted at least three times by the PEV-HLE. If the CM_PKCS_CERT.IND message is
    larger than 502 Octets, the message shall be fragmented by the HLE (refer to
    Section 11.1.7).
    """
    __hdr__ = (
        ("targetmac", "6s", b"\x00" * 6),
        ("cipersuite", "H", 0),
        ("certlen", "H", 0),
        ("certpackage", None, TriggerList)
    )

    targetmac_s = packetracer.get_property_mac("targetmac")


class CMPKCSCertRsp(Packet):
    __hdr__ = (
        ("targetmac", "6s", b"\x00" * 6),
        ("status", "B", 0),
        ("ciphersuitesize", "B", 0),  # optional
        ("cipersuite", None, TriggerList)
    )

    targetmac_s = packetracer.get_property_mac("targetmac")


MASK_FRAGINDEX = 0xF0
MASK_FRAGCOUNT = 0x0F


class Slac(Packet):
    __hdr__ = (
        ("version", "B", 1),
        ("typeinfo", "H", 0),  #
        ("frag_info", "B", 0),
        ("frag_seq", "B", 0)
    )

    __handler__ = {
        CM_SET_KEY | MMTYPELSB_REQUEST: CMSetKeyReq,
        CM_SET_KEY | MMTYPELSB_CONFIRM: CMSetKeyICnf,
        CM_SLAC_MATCH | MMTYPELSB_REQUEST: CMSlacMatchReq,
        CM_SLAC_MATCH | MMTYPELSB_CONFIRM: CMSlacMatchCnf,
        CM_ATTEN_CHAR | MMTYPELSB_INDICATION: CMAttenCharInd,
        CM_ATTEN_CHAR | MMTYPELSB_RESPONSE: CMAttenCharRsp,
        CM_SLAC_PARM | MMTYPELSB_REQUEST: CMSlacParmReq,
        CM_SLAC_PARM | MMTYPELSB_CONFIRM: CMSlacParmCnf,
        # MS:new
        CM_START_ATTEN_CHAR | MMTYPELSB_INDICATION: CMStartAttenCharInd,
        CM_MNBC_SOUND | MMTYPELSB_INDICATION: CMMnbcSoundInd,
        CM_LINK_STATS | MMTYPELSB_REQUEST: CMLinkStatsReq,
        CM_LINK_STATS | MMTYPELSB_CONFIRM: CMLinkStatsCnf,
        VS_PL_LNK_STATUS | MMTYPELSB_REQUEST: VSPLLinkStatusReq,
        VS_PL_LNK_STATUS | MMTYPELSB_CONFIRM: VSPLLinkStatusCnf,
        CM_AMP_MAP | MMTYPELSB_REQUEST: AMPMapReq,
        CM_AMP_MAP | MMTYPELSB_CONFIRM: AMPMapCnf,
        CM_PKCS_CERT | MMTYPELSB_REQUEST: CMPKCSCertReq,
        CM_PKCS_CERT | MMTYPELSB_CONFIRM: CMPKCSCertCnf,
        CM_PKCS_CERT | MMTYPELSB_INDICATION: CMPKCSCertInd,
        CM_PKCS_CERT | MMTYPELSB_RESPONSE: CMPKCSCertRsp
    }

    def _dissect(self, buf):
        typeinfo_be = unpack_H_le(buf[1: 3])[0]
        hlen = 5

        # VS_PL_LNK_STATUS does not have frag
        if typeinfo_be in {0xA0B8, 0xA0B9}:
            # logger.debug("disabling frag")
            self.frag_info = None
            self.frag_seq = None
            hlen = 3
        # logger.debug("Got type %X", typeinfo_be)
        self._init_handler(typeinfo_be, buf[hlen:])
        return hlen

    def _get_fragcount(self):
        return (self.frag_info & MASK_FRAGCOUNT)

    def _set_fragcount(self, fragcount):
        self.frag_info = (self.frag_info & ~MASK_FRAGCOUNT) | (fragcount & MASK_FRAGCOUNT)

    fragcount = property(_get_fragcount, _set_fragcount)

    def _get_fragindex(self):
        return (self.frag_info & MASK_FRAGINDEX) >> 4

    def _set_fragindex(self, fragindex):
        self.frag_info = (self.frag_info & ~MASK_FRAGINDEX) | ((fragindex << 4) & MASK_FRAGINDEX)

    fragindex = property(_get_fragindex, _set_fragindex)

    def _get_msgtype(self):
        typetmp = self.typeinfo & MASK_MSGTYPE_LE
        return unpack_H(pack_H_le(typetmp))[0]

    def _set_msgtype(self, msgtype):
        typetmp = unpack_H(pack_H_le(msgtype))[0]
        self.typeinfo = (self.typeinfo & ~MASK_MSGTYPE_LE) | (typetmp & MASK_MSGTYPE_LE)

    # base message type given as BE
    msgtype = property(_get_msgtype, _set_msgtype)

    def _get_msgtype_full(self):
        return unpack_H(pack_H_le(self.typeinfo))[0]

    def _set_msgtype_full(self, msgtype):
        self.typeinfo = unpack_H(pack_H_le(msgtype))[0]

    # set full typeinfo given as BE
    msgtype_full_be = property(_get_msgtype_full, _set_msgtype_full)

    def _get_msgtype_s(self):
        return TYPEINFO_DESCRIPTION.get(self.msgtype, None)

    msgtype_s = property(_get_msgtype_s)

    def _get_mmtypelsb(self):
        return (self.typeinfo & MASK_MMTYPELSB_LE) >> 8

    def _set_mmtypelsb(self, msgtype):
        typetmp = (self.typeinfo & ~MASK_MMTYPELSB_LE)
        self.typeinfo = typetmp | (msgtype << 8)

    # REQ->CNF, IND->RSP
    mmtypelsb = property(_get_mmtypelsb, _set_mmtypelsb)

    def _get_mmtypelsb_s(self):
        return MMTYPE_LSB_DESCRIPTION.get(self.mmtypelsb, None)

    mmtypelsb_s = property(_get_mmtypelsb_s, None)

    def _get_mmtypemsb(self):
        return (self.typeinfo & MASK_MMTYPEMSB_LE) >> 1

    def _set_mmtypemsb(self, msgtype):
        typetmp = (self.typeinfo & ~MASK_MMTYPEMSB_LE)
        self.typeinfo = typetmp | (msgtype << 1)

    mmtypemsb = property(_get_mmtypemsb, _set_mmtypemsb)

    def _get_mmtypemsb_s(self):
        return MMTYPE_MSB_DESCRIPTION.get(self.mmtypemsb, None)

    mmtypemsb_s = property(_get_mmtypemsb_s, None)



"""
Ethernet II, IEEE 802.3

RFC 1042
"""
import logging

from packetracer import packetracer, triggerlist
# handler
from packetracer.layer12 import arp, dtp, pppoe, flow_control, lacp
from packetracer.layer12 import lldp, slac
from packetracer.layer3 import ip, ip6, ipx
from packetracer.layer567 import ptpv2
from packetracer.packetracer_meta import FIELD_FLAG_IS_TYPEFIELD
from packetracer.structcbs import unpack_H

logger = logging.getLogger("packetracer")

ETH_CRC_LEN = 4
ETH_HDR_LEN = 14

ETH_LEN_MIN = 64  # minimum frame length with CRC
ETH_LEN_MAX = 1518  # maximum frame length with CRC

ETH_MTU = (ETH_LEN_MAX - ETH_HDR_LEN - ETH_CRC_LEN)
ETH_MIN = (ETH_LEN_MIN - ETH_HDR_LEN - ETH_CRC_LEN)

# Ethernet payload types - http://standards.ieee.org/regauth/ethertype
ETH_TYPE_PUP = 0x0200  # PUP protocol
ETH_TYPE_IP = 0x0800  # IPv4 protocol
ETH_TYPE_ARP = 0x0806  # address resolution protocol
ETH_TYPE_WOL = 0x0842  # Wake on LAN
ETH_TYPE_CDP = 0x2000  # Cisco Discovery Protocol
ETH_TYPE_DTP = 0x2004  # Cisco Dynamic Trunking Protocol
ETH_TYPE_REVARP = 0x8035  # reverse addr resolution protocol
ETH_TYPE_ETHTALK = 0x809B  # Apple Talk
ETH_TYPE_AARP = 0x80F3  # Appletalk Address Resolution Protocol
ETH_TYPE_8021Q = 0x8100  # IEEE 802.1Q VLAN tagging
ETH_TYPE_IPX = 0x8137  # Internetwork Packet Exchange
ETH_TYPE_NOV = 0x8138  # Novell
ETH_TYPE_IP6 = 0x86DD  # IPv6 protocol
ETH_TYPE_MPLS_UCAST = 0x8847  # MPLS unicast
ETH_TYPE_MPLS_MCAST = 0x8848  # MPLS multicast
ETH_TYPE_PPOE_DISC = 0x8863  # PPPoE Discovery
ETH_TYPE_PPOE_SESS = 0x8864  # PPPoE Session
ETH_TYPE_JUMBOF = 0x8870  # Jumbo Frames
ETH_TYPE_PROFINET = 0x8892  # Realtime-Ethernet PROFINET
ETH_TYPE_ATAOE = 0x88A2  # ATA other Ethernet
ETH_TYPE_ETHERCAT = 0x88A4  # Realtime-Ethernet Ethercat
ETH_TYPE_PBRIDGE = 0x88A8  # Provider Bridging IEEE 802.1ad
ETH_TYPE_POWERLINK = 0x88AB  # Realtime Ethernet POWERLINK
ETH_TYPE_LLDP = 0x88CC  # Link Layer Discovery Protocol
ETH_TYPE_SERCOS = 0x88CD  # Realtime Ethernet SERCOS III
ETH_TYPE_PTPv2 = 0x88F7  # PTPv2 IEEE 1588-2008
ETH_TYPE_FIBRE_ETH = 0x8906  # Fibre Channel over Ethernet
ETH_TYPE_FCOE = 0x8914  # FCoE Initialization Protocol (FIP)
ETH_TYPE_TUNNELING = 0x9100  # Provider Bridging IEEE 802.1QInQ 2007
ETH_TYPE_EFC = 0x8808  # Ethernet flow control
ETH_TYPE_SP = 0x8809  # Slow Protocols
ETH_TYPE_SLAC = 0x88E1  # SLAC

# MPLS label stack fields
MPLS_LABEL_MASK = 0xFFFFF000
MPLS_QOS_MASK = 0x00000E00
MPLS_TTL_MASK = 0x000000FF
MPLS_LABEL_SHIFT = 12
MPLS_QOS_SHIFT = 9
MPLS_TTL_SHIFT = 0
MPLS_STACK_BOTTOM = 0x0100


class Dot1Q(packetracer.Packet):
    __hdr__ = (
        ("type", "H", ETH_TYPE_8021Q),
        ("tci", "H", 0)  # tag control information PCP(3 bits),CFI(1 bit), VID(12 bits)
    )

    def __get_prio(self):
        return (self.tci & 0xE000) >> 13

    def __set_prio(self, value):
        self.tci = (self.tci & ~0xE000) | (value << 13)

    prio = property(__get_prio, __set_prio)

    def __get_cfi(self):
        return (self.tci & 0x1000) >> 12

    def __set_cfi(self, value):
        self.tci = (self.tci & ~0x1000) | (value << 12)

    cfi = property(__get_cfi, __set_cfi)

    def __get_vid(self):
        return self.tci & 0x0FFF

    def __set_vid(self, value):
        self.tci = self.tci & 0xF000 | value

    vid = property(__get_vid, __set_vid)

    type_t = packetracer.get_property_translator("type", "ETH_TYPE_")


# standard or double vlan tag
# ETH_TYPE_TUNNELING as outer tag is NON-standard!
# see: https://en.wikipedia.org/wiki/IEEE_802.1ad
VLAN_TAG_START = {ETH_TYPE_8021Q, ETH_TYPE_PBRIDGE, ETH_TYPE_TUNNELING}


class Ethernet(packetracer.Packet):
    __hdr__ = [
        ("dst", "6s", b"\xff" * 6),
        ("src", "6s", b"\xff" * 6),
        ("vlan", None, triggerlist.TriggerList),
        ("type", "H", ETH_TYPE_IP, FIELD_FLAG_IS_TYPEFIELD)
    ]

    dst_s = packetracer.get_property_mac("dst")
    src_s = packetracer.get_property_mac("src")
    type_t = packetracer.get_property_translator("type", "ETH_TYPE_")

    __handler__ = {
        ETH_TYPE_IP: ip.IP,
        ETH_TYPE_ARP: arp.ARP,
        ETH_TYPE_DTP: dtp.DTP,
        ETH_TYPE_IPX: ipx.IPX,
        ETH_TYPE_IP6: ip6.IP6,
        ETH_TYPE_PPOE_DISC: pppoe.PPPoE,
        ETH_TYPE_PPOE_SESS: pppoe.PPPoE,
        ETH_TYPE_PTPv2: ptpv2.PTPv2,
        ETH_TYPE_EFC: flow_control.FlowControl,
        ETH_TYPE_LLDP: lldp.LLDP,
        ETH_TYPE_SP: lacp.LACP,
        ETH_TYPE_SLAC: slac.Slac
    }

    def _dissect(self, buf):
        hlen = 14
        # Ethernet formats:
        # RFC 894 (Ethernet II) -> type = -> value >1500
        # 802.[2,3] (LLC format) -> type = length field -> value <=1500, not supported
        eth_type = unpack_H(buf[hlen - 2: hlen])[0]

        # any VLAN tag present? in this case: type field is actually a vlan tag
        if eth_type in VLAN_TAG_START:
            # TODO: use _init_triggerlist()
            if eth_type == ETH_TYPE_8021Q:
                # logger.debug(">>> got vlan tag")
                vlan_tag = Dot1Q(buf[12: 16])
                self.vlan.append(vlan_tag)
                hlen += 4
                # get real higher layer type
                eth_type = unpack_H(buf[16: 18])[0]
            # 802.1ad: support up to 2 tags (double tagging aka QinQ)
            else:
                # logger.debug(">>> got vlan tag")
                vlan_tag1 = Dot1Q(buf[12: 16])
                vlan_tag2 = Dot1Q(buf[16: 20])
                self.vlan.extend([vlan_tag1, vlan_tag2])
                hlen += 8
                # get real higher layer type
                eth_type = unpack_H(buf[20: 22])[0]

        # logger.debug("eth type is: %d" % eth_type)

        # handle ethernet-padding: remove it but save for later use
        # don't use headers for this because this is a rare situation
        dlen = len(buf) - hlen  # data length [+ padding?]

        # assume padding only present if len(higher_layer.bin()) <= 46
        if dlen <= 46:
            try:
                # this will only work on complete headers: Ethernet + IP + ...
                # handle padding using IPv4, IPv6 etc (min size "eth + ..." = 60 bytes)
                # logger.debug(">>> checking for padding")
                if eth_type == ETH_TYPE_IP:
                    dlen_ip = unpack_H(buf[hlen + 2: hlen + 4])[0]  # real data length

                    if dlen_ip < dlen:
                        # padding found
                        self._padding = buf[hlen + dlen_ip:]
                        # logger.debug("got padding for IPv4: %r" % self._padding)
                        dlen = dlen_ip
                # handle padding using IPv6
                # IPv6 is a piece of sh$§! payloadlength (in header) = exclusive standard header
                # but INCLUSIVE options!
                elif eth_type == ETH_TYPE_IP6:
                    dlen_ip = unpack_H(buf[hlen + 4: hlen + 6])[0]  # real data length
                    # logger.debug("eth.hlen=%d, data length based on header: %d" % (hlen, dlen_ip))

                    if 40 + dlen_ip < dlen:
                        # padding found
                        self._padding = buf[hlen + 40 + dlen_ip:]
                        # logger.debug("got padding for IPv6: %r" % self._padding)
                        dlen = 40 + dlen_ip
                elif eth_type == ETH_TYPE_LLDP:
                    # this is a bit redundant as we re-parse TLV when accessing the LLDP layer
                    dlen_lldp, _ = lldp.count_and_dissect_tlvs(buf[hlen:])
                    self._padding = buf[hlen + dlen_lldp:]
                    dlen = dlen_lldp
                elif eth_type == ETH_TYPE_SP:
                    lacppdu_len = 110
                    self._padding = buf[hlen + lacppdu_len:]
                    dlen = lacppdu_len
            except:
                # Could not extract padding info, assuming incomplete ethernet frame.
                # Init of handler will take place after all.
                pass
        # logger.debug("len(buf)=%d, len(higher)=%d" % (len(buf), dlen))
        self._init_handler(eth_type, buf[hlen: hlen + dlen])
        return hlen

    def _update_fields(self):
        self._update_higherlayer_id()

    def bin(self, update_auto_fields=True):
        # padding needs to be placed at the very end
        return packetracer.Packet.bin(self, update_auto_fields=update_auto_fields) + self.padding

    def __len__(self):
        return super().__len__() + len(self.padding)

    def direction(self, other):
        # logger.debug("checking direction: %s<->%s" % (self, other))
        if self.dst == other.dst and self.src == other.src:
            # consider packet to itself: can be DIR_REV
            return packetracer.Packet.DIR_SAME | packetracer.Packet.DIR_REV
        if (self.dst == other.src and self.src == other.dst) or \
                (self.dst == b"\xff\xff\xff\xff\xff\xff" and other.dst == self.src):  # broadcast
            return packetracer.Packet.DIR_REV
        return packetracer.Packet.DIR_UNKNOWN

    padding = packetracer.get_ondemand_property("padding", lambda: b"")

    def reverse_address(self):
        self.dst, self.src = self.src, self.dst



"""Internet Group Management Protocol."""

from packetracer import packetracer, checksum
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE


class IGMP(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
        ("maxresp", "B", 0),
        ("sum", "H", 0, FIELD_FLAG_AUTOUPDATE),
        ("group", "4s", b"\x00" * 4)
    )

    # Convenient access for: group[_s]
    group_s = packetracer.get_property_ip4("group")

    def _update_fields(self):
        if self.sum_au_active and self._changed():
            self.sum = 0
            self.sum = checksum.in_cksum(packetracer.Packet.bin(self))



"""Protocol Independent Multicast."""

from packetracer import packetracer, checksum
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE


class PIM(packetracer.Packet):
    __hdr__ = (
        ("v_type", "B", 0x20),
        ("rsvd", "B", 0),
        ("sum", "H", 0, FIELD_FLAG_AUTOUPDATE)  # _sum = sum
    )

    def __get_v(self):
        return self.v_type >> 4

    def __set_v(self, v):
        self.v_type = (v << 4) | (self.v_type & 0xF)

    v = property(__get_v, __set_v)

    def __get_type(self):
        return self.v_type & 0xF

    def __set_type(self, pimtype):
        self.v_type = (self.v_type & 0xF0) | pimtype

    type = property(__get_type, __set_type)

    def _update_fields(self):
        if self.sum_au_active and self._changed():
            self.sum = 0
            self.sum = checksum.in_cksum(packetracer.Packet.bin(self))



"""ISO Transport Service on top of the TCP (TPKT)."""

from packetracer import packetracer


# TPKT - RFC 1006 Section 6
# http://www.faqs.org/rfcs/rfc1006.html


class TPKT(packetracer.Packet):
    __hdr__ = (
        ("v", "B", 3),
        ("rsvd", "B", 0),
        ("len", "H", 0)
    )



"""Virtual Router Redundancy Protocol."""

from packetracer import packetracer, checksum
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE


class VRRP(packetracer.Packet):
    __hdr__ = (
        ("vtype", "B", 0x21),
        ("vrid", "B", 0),
        ("priority", "B", 0),
        ("count", "B", 0),
        ("atype", "B", 0),
        ("advtime", "B", 0),
        ("sum", "H", 0, FIELD_FLAG_AUTOUPDATE),
    )

    def __get_v(self):
        return self.vtype >> 4

    def __set_v(self, v):
        self.vtype = (self.vtype & ~0xF) | (v << 4)

    v = property(__get_v, __set_v)

    def __get_type(self):
        return self.vtype & 0xF

    def __set_type(self, v):
        self.vtype = (self.vtype & ~0xF0) | (v & 0xF)

    type = property(__get_type, __set_type)

    def _update_fields(self):
        if self.sum_au_active and self._changed():
            # logger.debug(">>> IP: calculating sum")
            # reset checksum for recalculation,  mark as changed / clear cache
            self.sum = 0
            # logger.debug(">>> IP: bytes for sum: %s" % self.header_bytes)
            self.sum = checksum.in_cksum(packetracer.Packet.bin(self, update_auto_fields=True))



"""
Internet Protocol version 4.

RFC 791
"""
import logging

from packetracer import packetracer, triggerlist, checksum
# handler
from packetracer.layer3 import esp, icmp, igmp, ip6, ipx, ospf, pim
from packetracer.layer3.ip_shared import IP_PROTO_IP6, IP_PROTO_ICMP, IP_PROTO_IGMP, IP_PROTO_TCP, \
    IP_PROTO_UDP, IP_PROTO_ESP, IP_PROTO_PIM, IP_PROTO_IPXIP, IP_PROTO_SCTP, IP_PROTO_OSPF
from packetracer.layer4 import tcp, udp, sctp
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE, FIELD_FLAG_IS_TYPEFIELD

logger = logging.getLogger("packetracer")

# avoid references for performance reasons
in_cksum = checksum.in_cksum

# IP options
# http://www.iana.org/assignments/ip-parameters/ip-parameters.xml
IP_OPT_EOOL = 0
IP_OPT_NOP = 1
IP_OPT_SEC = 2
IP_OPT_LSR = 3
IP_OPT_TS = 4
IP_OPT_ESEC = 5
IP_OPT_CIPSO = 6
IP_OPT_RR = 7
IP_OPT_SID = 8
IP_OPT_SSR = 9
IP_OPT_ZSU = 10
IP_OPT_MTUP = 11
IP_OPT_MTUR = 12
IP_OPT_FINN = 13
IP_OPT_VISA = 14
IP_OPT_ENCODE = 15
IP_OPT_IMITD = 16
IP_OPT_EIP = 17
IP_OPT_TR = 18
IP_OPT_ADDEXT = 19
IP_OPT_RTRALT = 20
IP_OPT_SDB = 21
IP_OPT_UNASSGNIED = 22
IP_OPT_DPS = 23
IP_OPT_UMP = 24
IP_OPT_QS = 25
IP_OPT_EXP = 30


class IPOptSingle(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
    )


class IPOptMulti(packetracer.Packet):
    """
    len = total length (header + data)
    """
    __hdr__ = (
        ("type", "B", 0),
        ("len", "B", 2),
    )

    def _update_fields(self):
        self.len = len(self)


class IP(packetracer.Packet):
    __hdr__ = (
        ("v_hl", "B", 69, FIELD_FLAG_AUTOUPDATE),  # = 0x45
        ("tos", "B", 0),
        ("len", "H", 20, FIELD_FLAG_AUTOUPDATE),
        ("id", "H", 0),
        ("frag_off", "H", 0),
        ("ttl", "B", 64),
        ("p", "B", IP_PROTO_TCP, FIELD_FLAG_IS_TYPEFIELD),
        ("sum", "H", 0, FIELD_FLAG_AUTOUPDATE),
        ("src", "4s", b"\x00" * 4),
        ("dst", "4s", b"\x00" * 4),
        ("opts", None, triggerlist.TriggerList)
    )

    __handler__ = {
        IP_PROTO_ICMP: icmp.ICMP,
        IP_PROTO_IGMP: igmp.IGMP,
        IP_PROTO_TCP: tcp.TCP,
        IP_PROTO_UDP: udp.UDP,
        IP_PROTO_IP6: ip6.IP6,
        IP_PROTO_ESP: esp.ESP,
        IP_PROTO_PIM: pim.PIM,
        IP_PROTO_IPXIP: ipx.IPX,
        IP_PROTO_SCTP: sctp.SCTP,
        IP_PROTO_OSPF: ospf.OSPF
    }

    UPDATE_DEPENDANTS = {tcp.TCP, udp.UDP}

    def __get_v(self):
        return self.v_hl >> 4

    def __set_v(self, value):
        self.v_hl = (value << 4) | (self.v_hl & 0xF)

    # version
    v = property(__get_v, __set_v)

    def __get_hl(self):
        return self.v_hl & 0x0F

    def __set_hl(self, value):
        self.v_hl = (self.v_hl & 0xF0) | value

    # header length
    hl = property(__get_hl, __set_hl)

    def __get_flags(self):
        return (self.frag_off & 0xE000) >> 13

    def __set_flags(self, value):
        self.frag_off = (self.frag_off & ~0xE000) | (value << 13)

    flags = property(__get_flags, __set_flags)

    def __get_offset(self):
        return self.frag_off & ~0xE000

    def __set_offset(self, value):
        self.frag_off = (self.frag_off & 0xE000) | value

    offset = property(__get_offset, __set_offset)

    def create_fragments(self, fragment_len=1480):
        """
        Create fragment packets from this IP packet with max fragment_len bytes each.
        This will set the flags and offset values accordingly (see header field off).

        fragment_len -- max length of a fragment (IP header + payload)
        return -- fragment IP packets created from this packet
        """
        if fragment_len % 8 != 0:
            raise Exception("fragment_len not multipe of 8 bytes: %r" % fragment_len)

        fragments = []
        length_ip_total = len(self.bin())
        payload = self.body_bytes
        length_ip_header = length_ip_total - len(payload)
        length_payload = length_ip_total - length_ip_header

        off = 0

        while off < length_payload:
            payload_sub = payload[off: off + fragment_len]

            ip_frag = IP(id=self.id, p=self.p, src=self.src, dst=self.dst)

            if length_payload - off > fragment_len:
                # more fragments follow
                ip_frag.flags = 0x1
            else:
                # last fragment
                ip_frag.flags = 0x0

            ip_frag.offset = int(off / 8)
            ip_frag.body_bytes = payload_sub
            fragments.append(ip_frag)
            off += fragment_len

        return fragments

    # Convenient access for: src[_s], dst[_s]
    src_s = packetracer.get_property_ip4("src")
    dst_s = packetracer.get_property_ip4("dst")
    p_t = packetracer.get_property_translator("p", "IP_PROTO_")

    def _dissect(self, buf):
        total_header_length = ((buf[0] & 0xF) << 2)
        options_length = total_header_length - 20  # total IHL - standard IP-len = options length

        if options_length > 0:
            # logger.debug("got some IP options: %s" % tl_opts)
            self._init_triggerlist("opts", buf[20: 20 + options_length], self._parse_opts)
        elif options_length < 0:
            # invalid header length: assume no options at all
            raise Exception()
        # TODO: extract real data length:
        # There are some cases where padding can not be identified on ethernet -> do it here (eg VSS shit trailer)
        self._init_handler(buf[9], buf[total_header_length:])
        return total_header_length

    __IP_OPT_SINGLE = {IP_OPT_EOOL, IP_OPT_NOP}

    @staticmethod
    def _parse_opts(buf):
        """Parse IP options and return them as list."""
        optlist = []
        i = 0
        p = None

        while i < len(buf):
            # logger.debug("got IP-option type %s" % buf[i])
            if buf[i] in IP.__IP_OPT_SINGLE:
                p = IPOptSingle(type=buf[i])
                i += 1
            else:
                olen = buf[i + 1]
                # logger.debug("IPOptMulti")
                p = IPOptMulti(type=buf[i], len=olen, body_bytes=buf[i + 2: i + olen])
                # logger.debug("body bytes: %s" % buf[i + 2: i + olen])
                i += olen  # typefield + lenfield + data-len
            # logger.debug("IPOptMulti 2")
            optlist.append(p)
        return optlist

    def _update_fields(self):
        self._update_higherlayer_id()

        if self.len_au_active:
            self.len = len(self)
        if self.v_hl_au_active:
            # Update header length. NOTE: needs to be a multiple of 4 Bytes.
            # logger.debug("updating: %r" % self._packet)
            # options length need to be multiple of 4 Bytes
            self.hl = int(self.header_len / 4) & 0xF
        if self.sum_au_active:
            # length changed so we have to recalculate checksum
            # logger.debug(">>> IP: calculating sum, current: %0X" % self.sum)
            # reset checksum for recalculation,  mark as changed / clear cache
            self.sum = 0
            # logger.debug(">>> IP: bytes for sum: %s" % self.header_bytes)
            self.sum = in_cksum(self._pack_header())
        # logger.debug("IP: new hl: %d / %d" % (self._packet.hdr_len, hdr_len_off))
        # logger.debug("new sum: %0X" % self.sum)

    def direction(self, other):
        # logger.debug("checking direction: %s<->%s" % (self, next))
        direction = 0
        if self.src == other.src and self.dst == other.dst:
            direction |= packetracer.Packet.DIR_SAME
        if self.src == other.dst and self.dst == other.src:
            direction |= packetracer.Packet.DIR_REV
        if direction == 0:
            direction = packetracer.Packet.DIR_UNKNOWN
        return direction

    def reverse_address(self):
        self.src, self.dst = self.dst, self.src


# Type of service (ip_tos), RFC 1349 ("obsoleted by RFC 2474")
IP_TOS_DEFAULT = 0x00  # default
IP_TOS_LOWDELAY = 0x10  # low delay
IP_TOS_THROUGHPUT = 0x08  # high throughput
IP_TOS_RELIABILITY = 0x04  # high reliability
IP_TOS_LOWCOST = 0x02  # low monetary cost - XXX
IP_TOS_ECT = 0x02  # ECN-capable transport
IP_TOS_CE = 0x01  # congestion experienced

# IP precedence (high 3 bits of ip_tos), hopefully unused
IP_TOS_PREC_ROUTINE = 0x00
IP_TOS_PREC_PRIORITY = 0x20
IP_TOS_PREC_IMMEDIATE = 0x40
IP_TOS_PREC_FLASH = 0x60
IP_TOS_PREC_FLASHOVERRIDE = 0x80
IP_TOS_PREC_CRITIC_ECP = 0xA0
IP_TOS_PREC_INTERNETCONTROL = 0xC0
IP_TOS_PREC_NETCONTROL = 0xE0

# Fragmentation flags (ip_off)
IP_RF = 0x4  # reserved
IP_DF = 0x2  # don't fragment
IP_MF = 0x1  # more fragments (not last frag)

# Time-to-live (ip_ttl), seconds
IP_TTL_DEFAULT = 64  # default ttl, RFC 1122, RFC 1340
IP_TTL_MAX = 255  # maximum ttl



"""
Transmission Control Protocol (TCP)

RFC 675 - Specification of Internet Transmission Control Program, December 1974 Version
RFC 793 - TCP v4
RFC 1122 - includes some error corrections for TCP
RFC 1323 - TCP-Extensions
RFC 1379 - Extending TCP for Transactions—Concepts
RFC 1948 - Defending Against Sequence Number Attacks
RFC 2018 - TCP Selective Acknowledgment Options
RFC 4614 - A Roadmap for TCP Specification Documents
RFC 5681 - TCP Congestion Control
RFC 6298 - Computing TCP's Retransmission Timer
RFC 6824 - TCP Extensions for Multipath Operation with Multiple Addresses
"""
import logging

from packetracer import packetracer, triggerlist, checksum
from packetracer.layer4 import ssl
from packetracer.layer567 import bgp, http, mqtt, rtp, sip, telnet, tpkt, pmap
from packetracer.packetracer_meta import FIELD_FLAG_AUTOUPDATE, FIELD_FLAG_IS_TYPEFIELD
from packetracer.structcbs import unpack_H, pack_ipv4_header, pack_ipv6_header

# avoid references for performance reasons
in_cksum = checksum.in_cksum

logger = logging.getLogger("packetracer")

# TCP control flags
TH_FIN = 0x01  # end of data
TH_SYN = 0x02  # synchronize sequence numbers
TH_RST = 0x04  # reset connection
TH_PUSH = 0x08  # push
TH_ACK = 0x10  # acknowledgment number set
TH_URG = 0x20  # urgent pointer set
TH_ECE = 0x40  # ECN echo, RFC 3168
TH_CWR = 0x80  # congestion window reduced

TCP_PORT_MAX = 65535  # maximum port
TCP_WIN_MAX = 65535  # maximum (unscaled) window

# TCP Options (opt_type) - http://www.iana.org/assignments/tcp-parameters
TCP_OPT_EOL = 0  # end of option list
TCP_OPT_NOP = 1  # no operation
TCP_OPT_MSS = 2  # maximum segment size
TCP_OPT_WSCALE = 3  # window scale factor, RFC 1072
TCP_OPT_SACKOK = 4  # SACK permitted, RFC 2018
TCP_OPT_SACK = 5  # SACK, RFC 2018
TCP_OPT_ECHO = 6  # echo (obsolete), RFC 1072
TCP_OPT_ECHOREPLY = 7  # echo reply (obsolete), RFC 1072
TCP_OPT_TIMESTAMP = 8  # timestamp, RFC 1323
TCP_OPT_POCONN = 9  # partial order conn, RFC 1693
TCP_OPT_POSVC = 10  # partial order service, RFC 1693
TCP_OPT_CC = 11  # connection count, RFC 1644
TCP_OPT_CCNEW = 12  # CC.NEW, RFC 1644
TCP_OPT_CCECHO = 13  # CC.ECHO, RFC 1644
TCP_OPT_ALTSUM = 14  # alt checksum request, RFC 1146
TCP_OPT_ALTSUMDATA = 15  # alt checksum data, RFC 1146
TCP_OPT_SKEETER = 16  # Skeeter
TCP_OPT_BUBBA = 17  # Bubba
TCP_OPT_TRAILSUM = 18  # trailer checksum
TCP_OPT_MD5 = 19  # MD5 signature, RFC 2385
TCP_OPT_SCPS = 20  # SCPS capabilities
TCP_OPT_SNACK = 21  # selective negative acks
TCP_OPT_REC = 22  # record boundaries
TCP_OPT_CORRUPT = 23  # corruption experienced
TCP_OPT_SNAP = 24  # SNAP
TCP_OPT_TCPCOMP = 26  # TCP compression filter
TCP_OPT_MAX = 27


class TCPOptSingle(packetracer.Packet):
    __hdr__ = (
        ("type", "B", 0),
    )

    type_t = packetracer.get_property_translator("type", "TCP_OPT_")


class TCPOptMulti(packetracer.Packet):
    """
    len = total length (header + data)
    """
    __hdr__ = (
        ("type", "B", 0),
        ("len", "B", 2, FIELD_FLAG_AUTOUPDATE)
    )

    type_t = packetracer.get_property_translator("type", "TCP_OPT_")

    def _update_fields(self):
        if self.len_au_active:
            self.len = len(self)


TCP_PROTO_TELNET = 23
TCP_PROTO_TPKT = 102
TCP_PROTO_PMAP = 111
TCP_PROTO_BGP = 179
TCP_PROTO_SSL = 443
TCP_PROTO_MQTT = (1883, 8883)
TCP_PROTO_HTTP = (80, 8008, 8080)
TCP_PROTO_RTP = (5004, 5005)
TCP_PROTO_SIP = (5060, 5061)


def cb_get_flag_description(value, value_name):
    descr = [name_d for value_d, name_d in value_name.items() if value_d & value != 0]
    return " | ".join(descr)


class TCP(packetracer.Packet):
    __hdr__ = (
        ("sport", "H", 0xDEAD),
        ("dport", "H", 0, FIELD_FLAG_IS_TYPEFIELD),
        ("seq", "I", 0xDEADBEEF),
        ("ack", "I", 0),
        ("off_x2", "B", ((5 << 4) | 0), FIELD_FLAG_AUTOUPDATE),  # 10*4 Byte
        ("flags", "B", TH_SYN),  # acces via (obj.flags & TH_XYZ)
        ("win", "H", TCP_WIN_MAX),
        ("sum", "H", 0, FIELD_FLAG_AUTOUPDATE),
        ("urp", "H", 0),
        ("opts", None, triggerlist.TriggerList)
    )

    # 4 bits | 4 bits
    # offset | reserved
    # offset * 4 = header length
    def __get_off(self):
        return self.off_x2 >> 4

    def __set_off(self, value):
        self.off_x2 = (value << 4) | (self.off_x2 & 0xF)

    off = property(__get_off, __set_off)

    # return real header length based on header info
    def __get_hlen(self):
        return self.off * 4

    # set real header length based on header info (should be n*4)
    def __set_hlen(self, value):
        self.off = int(value / 4)

    hlen = property(__get_hlen, __set_hlen)
    flags_t = packetracer.get_property_translator("flags", "TH_", cb_get_description=cb_get_flag_description)

    __handler__ = {
        TCP_PROTO_BGP: bgp.BGP,
        TCP_PROTO_TELNET: telnet.Telnet,
        TCP_PROTO_TPKT: tpkt.TPKT,
        TCP_PROTO_PMAP: pmap.Pmap,
        TCP_PROTO_MQTT: mqtt.MQTTBase,
        TCP_PROTO_HTTP: http.HTTP,
        TCP_PROTO_SSL: ssl.SSL,
        TCP_PROTO_RTP: rtp.RTP,
        TCP_PROTO_SIP: sip.SIP
    }

    def _update_fields(self):
        # TCP-checksum needs to be updated on one of the following:
        # - this layer itself or any higher layer changed
        # - changes to the IP-pseudoheader
        update = True
        # update header length. NOTE: needs to be a multiple of 4 Bytes.
        # options length need to be multiple of 4 Bytes
        if self._header_changed and self.off_x2_au_active:
            self.off = int(self.header_len / 4) & 0xF

        # we need some IP as lower layer
        if self._lower_layer is None:
            return

        # self._update_higherlayer_id()

        try:
            # changes to IP-layer, don't mind if this isn't IP
            if not self._lower_layer._header_changed:
                # pseudoheader didn't change, further check for changes in layers
                update = self._changed()
        # logger.debug("lower layer found!")
        except:
            # Assume not an IP packet: we can't calculate the checksum
            # logger.debug("no lower layer found!")
            update = False

        if update and self.sum_au_active:
            # logger.debug(">>> updating checksum")
            self._calc_sum()

    def _dissect(self, buf):
        # update dynamic header parts. buf: 1010???? -clear reserved-> 1010 -> *4
        ol = ((buf[12] >> 4) << 2) - 20  # dataoffset - TCP-standard length

        if ol > 0:
            # parse options, add offset-length to standard-length
            opts_bytes = buf[20: 20 + ol]
            self._init_triggerlist("opts", opts_bytes, self._parse_opts)
        elif ol < 0:
            raise Exception("invalid header length")

        ports = [unpack_H(buf[0:2])[0], unpack_H(buf[2:4])[0]]

        try:
            # source or destination port should match
            # logger.debug("TCP handler: %r" % self._id_handlerclass_dct[TCP])
            htype = [x for x in ports if x in self._id_handlerclass_dct[TCP]][0]
            # logger.debug("TCP: trying to set handler, type: %d = %s" %
            # (type, self._id_handlerclass_dct[TCP][type]))
            self._init_handler(htype, buf[20 + ol:])
        except:
            # no type found
            pass
        return 20 + ol

    __TCP_OPT_SINGLE = {TCP_OPT_EOL, TCP_OPT_NOP}

    @staticmethod
    def _parse_opts(buf):
        """Parse TCP options using buf and return them as List."""
        optlist = []
        i = 0

        while i < len(buf):
            # logger.debug("got TCP-option type %s" % buf[i])
            if buf[i] in TCP.__TCP_OPT_SINGLE:
                p = TCPOptSingle(type=buf[i])
                i += 1
            else:
                olen = buf[i + 1]
                # p = TCPOptMulti(type=buf[i], len=olen, body_bytes=buf[i + 2: i + olen])
                p = TCPOptMulti(buf[i: i + olen])
                i += olen  # typefield + lenfield + data-len
            optlist.append(p)
        # logger.debug("tcp: parseopts finished, length: %d" % len(optlist))
        return optlist

    def _calc_sum(self):
        """Recalculate the TCP-checksum. This won't reset changed state."""
        # TCP and underwriting are freaky bitches: we need the IP pseudoheader
        # to calculate their checksum.
        try:
            # we need src/dst for checksum-calculation
            src, dst = self._lower_layer.src, self._lower_layer.dst
            self.sum = 0
            # logger.debug("TCP sum recalc: IP=%d / %s / %s" % (len(src), src, dst))

            tcp_bin = self.header_bytes + self.body_bytes
            # IP-pseudoheader, check if version 4 or 6
            if len(src) == 4:
                s = pack_ipv4_header(src, dst, 6, len(tcp_bin))  # 6 = TCP
            else:
                s = pack_ipv6_header(src, dst, 6, len(tcp_bin))  # 6 = TCP

            # Get checksum of concatenated pseudoheader+TCP packet
            # logger.debug("pseudoheader: %r" % s)
            # logger.debug("tcp_bin: %r" % tcp_bin)
            # assign via non-shadowed variable to trigger re-packing
            self.sum = in_cksum(s + tcp_bin)
        # logger.debug(">>> new checksum: %0X" % self._sum)
        except:
            # Not an IP packet as lower layer (src, dst not present) or invalid src/dst
            # logger.debug("could not calculate checksum: %r" % e)
            pass

    def direction(self, other):
        direction = 0
        # logger.debug("checking direction: %s<->%s" % (self, other))
        if self.sport == other.sport and self.dport == other.dport:
            direction |= packetracer.Packet.DIR_SAME
        if self.sport == other.dport and self.dport == other.sport:
            direction |= packetracer.Packet.DIR_REV
        if direction == 0:
            direction = packetracer.Packet.DIR_UNKNOWN
        return direction

    def reverse_address(self):
        self.sport, self.dport = self.dport, self.sport

    ra_segments = packetracer.get_ondemand_property("ra_segments", lambda: {})

    def ra_collect(self, pkt_list):
        """
        Collect TCP segments which have the same direction as this packet.
        Concatenated segments can be retrieved via ra_bin(). Does not check for missing
        segments.
        return -- bytes_cnt, [True|False]: amount of bytes added (sum of body bytes),
            final packet found (RST or FIN). True indicates that ra_bin() can be called.
        """
        if type(pkt_list) is not list:
            pkt_list = [pkt_list]

        bts_cnt = 0

        for segment in pkt_list:
            if self.direction(segment) != packetracer.Packet.DIR_SAME or len(segment.body_bytes) == 0:
                continue

            seq_store = segment.seq
            # final packet found: connection is going to be terminated
            if (segment.flags & TH_FIN) != 0 or (segment.flags & TH_RST) != 0:
                return 0, True

            if seq_store < self.seq:
                logger.warning("seq of new segment is lower than start")
                seq_store += 0xFFFF

            # logger.debug("adding tcp segment: %r", segment.body_bytes)
            self.ra_segments[seq_store] = segment.body_bytes
            bts_cnt += len(segment.body_bytes)

        return bts_cnt, False

    def ra_bin(self):
        """
        Assemble retrieved TCP segments in sorted order (body bytes of TCP segments).
        Does NOT flush the internal buffer; this is done via self.ra_segments.clear()
        """
        self.ra_segments[self.seq] = self.body_bytes
        sorted_list = sorted(self.ra_segments.items(), key=lambda t: t[0])
        bts_lst = [value for key, value in sorted_list]
        self.ra_segments.clear()
        return b"".join(bts_lst)



"""
Diameter Base Protocol - RFC 3588
http://tools.ietf.org/html/rfc3588
"""
import logging

from packetracer import packetracer, triggerlist

logger = logging.getLogger("packetracer")

# Request/Answer Command Codes
ABORT_SESSION = 274
ACCOUTING = 271
CAPABILITIES_EXCHANGE = 257
DEVICE_WATCHDOG = 280
DISCONNECT_PEER = 282
RE_AUTH = 258
SESSION_TERMINATION = 275


class Diameter(packetracer.Packet):
    __hdr__ = (
        ("v", "B", 1),
        ("len", "3s", b"\x00" * 3),
        ("flags", "B", 0),
        ("cmd", "3s", b"\x00" * 3),
        ("app_id", "I", 0),
        ("hop_id", "I", 0),
        ("end_id", "I", 0),
        ("avps", None, triggerlist.TriggerList)
    )

    def __get_r(self):
        return (self.flags >> 7) & 0x1

    def __set_r(self, r):
        self.flags = (self.flags & ~0x80) | ((r & 0x1) << 7)

    request_flag = property(__get_r, __set_r)

    def __get_p(self):
        return (self.flags >> 6) & 0x1

    def __set_p(self, p):
        self.flags = (self.flags & ~0x40) | ((p & 0x1) << 6)

    proxiable_flag = property(__get_p, __set_p)

    def __get_e(self):
        return (self.flags >> 5) & 0x1

    def __set_e(self, e):
        self.flags = (self.flags & ~0x20) | ((e & 0x1) << 5)

    error_flag = property(__get_e, __set_e)

    def __get_t(self):
        return (self.flags >> 4) & 0x1

    def __set_t(self, t):
        self.flags = (self.flags & ~0x10) | ((t & 0x1) << 4)

    retransmit_flag = property(__get_t, __set_t)

    def _dissect(self, buf):
        self._init_triggerlist("avps", buf[20:], self._parse_avps)
        return len(buf)

    def _parse_avps(self, buf):
        off = 0
        avps = []
        buflen = len(buf)

        # parse AVPs
        while off < buflen:
            avplen = int.from_bytes(buf[off + 5: off + 8], "big")
            # REAL length of AVP is multiple of 4 Bytes
            mod_len = avplen % 4
            if mod_len != 0:
                avplen += 4 - mod_len
            avp = AVP(buf[off: off + avplen])
            avps.append(avp)
            off += avplen
        return avps


class AVP(packetracer.Packet):
    __hdr__ = (
        ("code", "I", 0),
        ("flags", "B", 0),
        ("len", "3s", b""),
    )

    def __get_v(self):
        return (self.flags >> 7) & 0x1

    def __set_v(self, v):
        self.flags = (self.flags & ~0x80) | ((v & 0x1) << 7)

    vendor_flag = property(__get_v, __set_v)

    def __get_m(self):
        return (self.flags >> 6) & 0x1

    def __set_m(self, m):
        self.flags = (self.flags & ~0x40) | ((m & 0x1) << 6)

    mandatory_flag = property(__get_m, __set_m)

    def __get_p(self):
        return (self.flags >> 5) & 0x1

    def __set_p(self, p):
        self.flags = (self.flags & ~0x20) | ((p & 0x1) << 5)

    protected_flag = property(__get_p, __set_p)



"""PPP-over-Ethernet."""
from packetracer import packetracer
from packetracer.layer12.ppp import PPP

# RFC 2516 codes
PPPoE_PADI = 0x09
PPPoE_PADO = 0x07
PPPoE_PADR = 0x19
PPPoE_PADS = 0x65
PPPoE_PADT = 0xA7
PPPoE_SESSION = 0x00


class PPPoE(packetracer.Packet):
    __hdr__ = (
        ("v_type", "B", 0x11),
        ("code", "B", 0),
        ("session", "H", 0),
        ("len", "H", 0)  # payload length
    )

    def __get_v(self):
        return self.v_type >> 4

    def __set_v(self, v):
        self.v_type = (v << 4) | (self.v_type & 0xF)

    v = property(__get_v, __set_v)

    def __get_type(self):
        return self.v_type & 0xF

    def __set_type(self, t):
        self.v_type = (self.v_type & 0xF0) | t

    type = property(__get_type, __set_type)

    def _dissect(self, buf):
        code = buf[1]
        if code == PPPoE_SESSION:
            try:
                self._set_bodyhandler(PPP(buf[6:]))
            except Exception:
                pass
        else:
            pass
        return 6



"""
Bluetooth Low Energy

https://www.bluetooth.com/specifications/adopted-specifications
https://developer.bluetooth.org/TechnologyOverview/Pages/BLE.aspx
https://developer.bluetooth.org/TechnologyOverview/Pages/LE-Security.aspx
"""
import logging
from binascii import hexlify

from packetracer import packetracer
from packetracer import triggerlist
from packetracer.checksum import crc_btle_check
from packetracer.structcbs import unpack_I

logger = logging.getLogger("packetracer")

"""
flags as BE (in packet: as LE), 0x0001 becomes 0x0100
0x0001 indicates the LE Packet is de-whitened
0x0002 indicates the Signal Power field is valid
0x0004 indicates the Noise Power field is valid
0x0008 indicates the LE Packet is decrypted
0x0010 indicates the Reference Access Address is valid and led to this packet being captured
0x0020 indicates the Access Address Offenses field contains valid data
0x0040 indicates the RF Channel field is subject to aliasing
0x0400 indicates the CRC portion of the LE Packet was checked
0x0800 indicates the CRC portion of the LE Packet passed its check
0x1000 indicates the MIC portion of the decrypted LE Packet was checked
0x2000 indicates the MIC portion of the decrypted LE Packet passed its check
"""

_WHITE_MASK = (0x0100, 8)
_SIG_MASK = (0x0200, 9)
_NOISE_MASK = (0x0400, 10)
_DECR_MASK = (0x0800, 11)
_REF_ACC_MASK = (0x1000, 12)
_OFFENSE_MASK = (0x2000, 13)
_CHAN_ALIAS_MASK = (0x4000, 14)
_CRC_CHECK_MASK = (0x0004, 2)
_CRC_PASS_MASK = (0x0008, 3)
_MIC_CHECK_MASK = (0x0010, 4)
_MIC_PASS_MASK = (0x0020, 5)

FLAGS_NAME_MASK = {
    "whitening": _WHITE_MASK,
    "sigvalid": _SIG_MASK,
    "noisevalid": _NOISE_MASK,
    "decrypted": _DECR_MASK,
    "refaavalid": _REF_ACC_MASK,
    "aaoffensesvalid": _OFFENSE_MASK,
    "chanalias": _CHAN_ALIAS_MASK,
    "crcchecked": _CRC_CHECK_MASK,
    "crcvalid": _CRC_PASS_MASK,
    "micchecked": _MIC_CHECK_MASK,
    "micvalid": _MIC_PASS_MASK
}

BTLE_HANDLE_TYPE = 0
_subheader_properties = []

str_upper = str.upper
bytes_decode = bytes.decode


def reverse_bts(bts):
    return bytes(reversed(bytearray(bts)))


def reverse_bts_to_str(bts):
    bts_rev = reverse_bts(bts)
    return str_upper(bytes_decode(hexlify(bts_rev)))


# Set properties to access flags
for name, mask_off in FLAGS_NAME_MASK.items():
    subheader = [
        name,
        (lambda mask, off: (lambda _obj: (_obj.flags & mask) >> off))(mask_off[0], mask_off[1]),
        (lambda mask, off: (lambda _obj, _val: _obj.__setattr__("flags", (_obj.flags & ~mask) | (_val << off)))
         )(mask_off[0], mask_off[1])
    ]
    _subheader_properties.append(subheader)


class AdvData(packetracer.Packet):
    __hdr__ = (
        ("len", "B", 0),
        ("type", "B", 0)
    )


def parse_advdata(bts):
    off = 0
    ret = []

    while off < len(bts):
        alen = bts[off]
        pkt = AdvData(len=alen, type=bts[off + 1], body_bytes=bts[off + 2: off + 1 + alen])
        ret.append(pkt)
        off += 1 + alen
    return ret


#
# Sub header
#


class AdvInd(packetracer.Packet):
    __hdr__ = (
        ("adv_addr", "6s", b"\xff" * 6),
        ("adv_data", None, triggerlist.TriggerList),
    )

    def _dissect(self, buf):
        self._init_triggerlist("adv_data", buf[6:], parse_advdata)
        return len(buf)

    adv_addr_s = property(lambda obj: reverse_bts_to_str(obj.adv_addr))
    adv_data_s = property(lambda obj: reverse_bts_to_str(obj.adv_data))


class AdvNonconnInd(packetracer.Packet):
    __hdr__ = (
        ("adv_addr", "6s", b"\xff" * 6),
        ("adv_data", None, triggerlist.TriggerList)
    )

    def _dissect(self, buf):
        self._init_triggerlist("adv_data", buf[6:], parse_advdata)
        return len(buf)

    adv_addr_s = property(lambda obj: reverse_bts_to_str(obj.adv_addr))
    adv_data_s = property(lambda obj: reverse_bts_to_str(obj.adv_data))


class ScanRequest(packetracer.Packet):
    __hdr__ = (
        ("scan_addr", "6s", b"\xff" * 6),
        ("adv_addr", "6s", b"\xff" * 6),
    )

    scan_addr_s = property(lambda obj: reverse_bts_to_str(obj.scan_addr))
    adv_addr_s = property(lambda obj: reverse_bts_to_str(obj.adv_addr))


class ScanResponse(packetracer.Packet):
    __hdr__ = (
        ("adv_addr", "6s", b"\xff" * 6),
        ("adv_data", None, triggerlist.TriggerList)
    )

    def _dissect(self, buf):
        self._init_triggerlist("adv_data", buf[6:], parse_advdata)
        return len(buf)

    adv_addr_s = property(lambda obj: reverse_bts_to_str(obj.adv_addr))
    adv_data_s = property(lambda obj: reverse_bts_to_str(obj.adv_data))


class ConnRequest(packetracer.Packet):
    __hdr__ = (
        ("init_addr", "6s", b"\xff" * 6),
        ("adv_addr", "6s", b"\xff" * 6),
        ("access_addr", "4s", b"\xff" * 6),
        ("crcinit", "3s", b"\xff" * 3),
        ("winsize", "B", 0),
        ("winoff", "H", 0),
        ("interval", "H", 0),
        ("latency", "H", 0),
        ("timeout", "H", 0),
        ("chanmap", "5s", b"\xff" * 5),
        ("hop_sleep", "B", 0)
    )

    init_addr_s = property(lambda obj: reverse_bts_to_str(obj.init_addr))
    adv_addr_s = property(lambda obj: reverse_bts_to_str(obj.adv_addr))
    access_addr_s = property(lambda obj: reverse_bts_to_str(obj.access_addr))
    crcinit_s = property(lambda obj: reverse_bts_to_str(obj.crcinit))
    crcinit_rev = property(lambda obj: reverse_bts(obj.crcinit))

    def get_active_channels(self):
        """
        return -- Complete mapping of all channels as list
            including mapping incactive -> active like (1, 2, 3, ...)
        """
        data_channels_ret = []
        data_channels_active = []
        channel_current = 0
        active_total = 0

        for bt in self.chanmap:
            bit = 1
            for _ in range(8):
                if bt & bit != 0:
                    data_channels_ret.append(channel_current)
                    data_channels_active.append(channel_current)
                    active_total += 1
                else:
                    data_channels_ret.append(None)
                bit <<= 1
                channel_current += 1

                if channel_current >= 37:
                    break

        for idx, channel in enumerate(data_channels_ret):
            if channel is None:
                data_channels_ret[idx] = data_channels_active[idx % active_total]
        return data_channels_ret

    def __get_crcinit_int(self):
        return unpack_I(b"\x00" + self.crcinit)[0]

    crcinit_int = property(__get_crcinit_int)

    def __get_crcinit_rev_int(self):
        return unpack_I(b"\x00" + self.crcinit_rev)[0]

    crcinit_rev_int = property(__get_crcinit_rev_int)


#
# Data packets
#


class DataLLID0(packetracer.Packet):
    pass


class DataLLID1(packetracer.Packet):
    pass


class DataLLID2(packetracer.Packet):
    pass


# LLX-packets
class LLTerminateInd(packetracer.Packet):
    pass


class LLEncReq(packetracer.Packet):
    __hdr__ = (
        ("rand", "8s", b"\x00" * 8),
        ("encrdiv", "H", 0),
        ("masterdiv", "Q", 0),
        ("masterinit", "I", 0)
    )


class LLEncResp(packetracer.Packet):
    __hdr__ = (
        ("slavediv", "8s", b"\x00" * 8),
        ("slaveinit", "I", 0)
    )


class LLStartEnc(packetracer.Packet):
    pass


class LLVersionInd(packetracer.Packet):
    __hdr__ = (
        ("version", "B", 0),
        ("company", "H", 0),
        ("subcompany", "H", 0)
    )


class LLFeatureReq(packetracer.Packet):
    pass


class LLRejectInd(packetracer.Packet):
    pass


LLID3_TERMINATEIND = 0x2
LLID3_ENCREQ = 0x3
LLID3_ENCRESP = 0x4
LLID3_STARTENC = 0x5
LLID3_FEATUREREQ = 0x8
LLID3_VERSIONIND = 0xC
LLID3_REJECTIND = 0xD


class DataLLID3(packetracer.Packet):
    __hdr__ = (
        ("opcode", "B", 0),
    )

    __handler__ = {
        LLID3_TERMINATEIND: LLTerminateInd,
        LLID3_ENCREQ: LLEncReq,
        LLID3_ENCRESP: LLEncResp,
        LLID3_STARTENC: LLStartEnc,
        LLID3_VERSIONIND: LLVersionInd,
        LLID3_FEATUREREQ: LLFeatureReq,
        LLID3_REJECTIND: LLRejectInd
    }

    def _dissect(self, buf):
        self._init_handler(buf[0], buf[1:])
        return 1


#
# Base header
#
# Spec 4.0, p2203

PDU_TYPE_ADV_IND = 0
PDU_TYPE_ADV_DIRECT_IND = 1
PDU_TYPE_ADV_NONCONN_IND = 2
PDU_TYPE_SCAN_REQ = 3
PDU_TYPE_SCAN_RSP = 4
PDU_TYPE_CONNECT_REQ = 5
PDU_TYPE_ADV_SCAN_IND = 6
# 0111-1111 Reserved

# unknown
PDU_TYPE_DATA_LLID0 = 0
# Cont frag of L2CAP msg or empty PDU
PDU_TYPE_DATA_LLID1 = 1
# Start of L2CAP msg or complete L2CAP w/o frag
PDU_TYPE_DATA_LLID2 = 2
# Ctrl PDU
PDU_TYPE_DATA_LLID3 = 3


def _get_property_subtype_get(obj):
    if obj.access_addr == b"\xd6\xbe\x89\x8e":
        return obj.info & 0x0F
    return obj.info & 0x03


def _get_property_subtype_set(obj, val):
    if obj.access_addr == b"\xd6\xbe\x89\x8e":
        obj.info = (obj.info & ~0x0F) | val
    else:
        obj.info = (obj.info & ~0x03) | val


_subheader_btle_properties = [
    ["pdutype",
     lambda _obj: _get_property_subtype_get(_obj),
     lambda _obj, _val: _get_property_subtype_set(_obj, _val)],
    ["random_rx",
     lambda _obj: (_obj.info & 0x80) >> 7,
     lambda _obj, _val: (_obj.info & ~0x80) | (_val << 7)],
    ["random_tx",
     lambda _obj: (_obj.info & 0x40) >> 6,
     lambda _obj, _val: (_obj.info & ~0x40) | (_val << 6)],
    ["llid",
     lambda _obj: _obj.info & 0x03,
     lambda _obj, _val: (_obj.info & ~0x03) | _val],
    ["is_adv",
     lambda _obj: _obj.access_addr == b"\xd6\xbe\x89\x8e"]
]


class BTLE(packetracer.Packet):
    __hdr__ = (
        ("access_addr", "4s", b"\xff" * 4),
        ("info", "B", 0),
        ("len", "B", 0),
    )

    __hdr_sub__ = _subheader_btle_properties

    __handler__ = {
        PDU_TYPE_ADV_IND: AdvInd,
        PDU_TYPE_ADV_SCAN_IND: ScanRequest,
        PDU_TYPE_ADV_NONCONN_IND: AdvNonconnInd,
        PDU_TYPE_SCAN_REQ: ScanRequest,
        PDU_TYPE_SCAN_RSP: ScanResponse,
        PDU_TYPE_CONNECT_REQ: ConnRequest,
        (PDU_TYPE_DATA_LLID0 + 1) << 8: DataLLID0,
        (PDU_TYPE_DATA_LLID1 + 1) << 8: DataLLID1,
        (PDU_TYPE_DATA_LLID2 + 1) << 8: DataLLID2,
        (PDU_TYPE_DATA_LLID3 + 1) << 8: DataLLID3,
    }

    def _dissect(self, buf):
        hlen = 6
        # logger.debug("buf: %r" % buf)

        if buf[: 4] == b"\xd6\xbe\x89\x8e":
            # logger.debug("got ADV... packet")
            btle_type = buf[4] & 0x0F
        else:
            # logger.debug("got data packet")
            # max value is 15, shift to avoid collision with ADV... packets
            btle_type = ((buf[4] & 0x03) + 1) << 8
        # logger.warning("unpacked type: %r" % btle_type)
        self._init_handler(btle_type, buf[hlen: -3])

        self._crc = buf[-3:]
        return hlen

    def bin(self, update_auto_fields=True):
        """Custom bin(): handle crc for BTLE."""
        return packetracer.Packet.bin(self, update_auto_fields=update_auto_fields) + self.crc

    def __len__(self):
        return super().__len__() + len(self.crc)

    # handle crc attribute
    def __get_crc(self):
        try:
            return self._crc
        except AttributeError:
            return b""

    def __set_crc(self, crc):
        self._crc = crc

    crc = property(__get_crc, __set_crc)

    def is_crc_ok(self, crc_init=0xAAAAAA):
        return crc_btle_check(self.bin(), crc_init)

    crc_ok = property(is_crc_ok)


# BTLE packet header
# http://www.tcpdump.org/linktypes.html -> LINKTYPE_BLUETOOTH_LE_LL_WITH_PHDR
class BTLEHdr(packetracer.Packet):
    __hdr__ = (
        ("channel", "B", 0),
        ("signal", "B", 0),
        ("noise", "B", 0),
        ("aaoffense", "B", 0),
        ("refaddr", "4s", b"\xff" * 4),
        ("flags", "H", 0),
    )

    __hdr_sub__ = _subheader_properties

    __handler__ = {
        BTLE_HANDLE_TYPE: BTLE
    }

    def _dissect(self, buf):
        self._init_handler(BTLE_HANDLE_TYPE, buf[10:])
        # logger.debug("BTLE header: %r" % buf[:10])
        # logger.debug(adding %d flags" % len(self.flags))
        return 10



"""Server Message Block."""

from packetracer import packetracer


class SMB(packetracer.Packet):
    __hdr__ = [
        ("proto", "4s", b""),
        ("cmd", "B", 0),
        ("err", "I", 0),
        ("flags1", "B", 0),
        ("flags2", "B", 0),
        ("pad", "6s", b""),
        ("tid", "H", 0),
        ("pid", "H", 0),
        ("uid", "H", 0),
        ("mid", "H", 0)
    ]



"""
Controller Area Network, ISO 11898-1, see https://en.wikipedia.org/wiki/CAN_bus
SocketCAN message format handling

API examples:
https://github.com/rhyttr/SocketCAN/tree/3c46872d9af0885b42526b70853400c6d94b7c54/can-utils
"""
import logging
import sys

from packetracer import packetracer

logger = logging.getLogger("packetracer")

module_this = sys.modules[__name__]

"""
OBDII modes:

01 	Show current data
02 	Show freeze frame data
03 	Show stored Diagnostic Trouble Codes
04 	Clear Diagnostic Trouble Codes and stored values
05 	Test results, oxygen sensor monitoring (non CAN only)
06 	Test results, other component/system monitoring
07 	Show pending Diagnostic Trouble Codes (detected during current or last driving cycle)
08 	Control operation of on-board component/system
09 	Request vehicle information
0A 	Permanent Diagnostic Trouble Codes (DTCs) (Cleared DTCs)
"""

OBD2_MODE_01_PIDS = {
    0x00: "PIDS_SUPPORTED_1_20",
    0x01: "MONITOR_STATUS",
    0x02: "FREEZE_DTC",
    0x03: "FUEL_SYSTEM_STATUS",
    0x04: "CALCULATED_ENGINE_LOAD",
    0x05: "ENGINE_COOLANT_TEMP",
    0x06: "SHORT_TERM_FUEL_BANK1",
    0x07: "LONG_TERM_FUEL_BANK1",
    0x08: "SHORT_TERM_FUEL_BANK2",
    0x09: "LONG_TERM_FUEK_BANK2",
    0x0A: "FUELPRESSURE",
    0x0B: "INTAKE_MANIFOLD_PRESSURE",
    0x0C: "ENGINE_RPM",
    0x0D: "VEHICLE_SPEED",
    0x0E: "TIMING_ADVANCE",
    0x0F: "INTAKGE_AIR_TEMP",
    0x10: "MAF_AIR_FLOW",
    0x11: "THROTTLE_POS",
    0x12: "COMMANDED_SECONDARY_AIR_STATUS",
    0x13: "OXYGEN_SENSORS_PRESENT",
    0x14: "OXYGEN_SENSOR1",
    0x15: "OXYGEN_SENSOR2",
    0x16: "OXYGEN_SENSOR3",
    0x17: "OXYGEN_SENSOR4",
    0x18: "OXYGEN_SENSOR5",
    0x19: "OXYGEN_SENSOR6",
    0x1A: "OXYGEN_SENSOR7",
    0x1B: "OXYGEN_SENSOR8",
    0x1C: "OBD_STANDARDS_",
    0x1D: "OXYGEN_SENSORS",
    0x1E: "AUXILIARY_INPUT_STATUS",
    0x1F: "RUNTIME_SINCE_ENGINE",
    0x20: "PIDS_SUPPORTED_21_40",
    0x21: "DISTANCE_TRAVELED_WITH_MALFUNCTION",
    0x22: "FUEL_RAIL_PRESSURE",
    0x23: "FUEL_RAIL_GAUGE_PRESSURE",
    0x24: "OXYGEN_SENSOR1",
    0x25: "OXYGEN_SENSOR2",
    0x26: "OXYGEN_SENSOR3",
    0x27: "OXYGEN_SENSOR4",
    0x28: "OXYGEN_SENSOR5",
    0x29: "OXYGEN_SENSOR6",
    0x2A: "OXYGEN_SENSOR7",
    0x2B: "OXYGEN_SENSOR8",
    0x2C: "COMMANDED_EGR",
    0x2D: "EGR_ERROR",
    0x2E: "COMMANDED_EVAPORATIVE_PURGE",
    0x2F: "FUEL_TANK_LEVEL_INPUT",
    0x30: "WARMUPS_SINCE_CODES_CLEARED",
    0x31: "DISTANCE_TRAVELED_SINCE_CODES_CLEARED",
    0x32: "EVAP_SYSTEM_VAPOR_PRESSURE",
    0x33: "ABSOLUTE_BAROMETIC_PRESSURE",
    0x34: "OXYGEN_SENSOR1_FUELAIR",
    0x35: "OXYGEN_SENSOR2_FUELAIR",
    0x36: "OXYGEN_SENSOR3_FUELAIR",
    0x37: "OXYGEN_SENSOR4_FUELAIR",
    0x38: "OXYGEN_SENSOR5_FUELAIR",
    0x39: "OXYGEN_SENSOR6_FUELAIR",
    0x3A: "OXYGEN_SENSOR7_FUELAIR",
    0x3B: "OXYGEN_SENSOR8_FUELAIR",
    0x3C: "CATALYST_TEMP_B1S1",
    0x3D: "CATALYST_TEMP_B2S1",
    0x3E: "CATALYST_TEMP_B1S2",
    0x3F: "CATALYST_TEMP_B2S2",
    0x40: "PIDS_SUPPORTED_41_60",
    0x41: "MONITOR_STATUS_",
    0x42: "CTRL_MODULE_VOLTAGE",
    0x43: "ABS_LOAD_VALUE",
    0x44: "FUEL_AIR_COMMANDED",
    0x45: "RELATIVE_THROTTLE_POS",
    0x46: "AMBIENT_AIR_TEMP",
    0x47: "ABSOLUTE_THROTTLE_POS_B",
    0x48: "ABSOLUTE_THROTTLE_POS_C",
    0x49: "ABSOLUTE_THROTTLE_POS_D",
    0x4A: "ABSOLUTE_THROTTLE_POS_E",
    0x4B: "ABSOLUTE_THROTTLE_POS_F",
    0x4C: "COMMANDED_THROTTLE_ACTUATOR",
    0x4D: "TIME_RUN_WITH_MIL",
    0x4E: "TIME_SINCE_TROUBLE_CODES_CLEARED",
    0x4F: "MAX_FUEL_AIR_EQ_RATIO",
    0x50: "MAX_AIR_FLOW",
    0x51: "FUEL_TYPE",
    0x52: "ETHANOL_FUEL",
    0x53: "ABS_EVAP_SYSTEM_VAPOR_PRESSURE",
    0x54: "EVAP_SYSTEM_VAPOR_PRESSURE",
    0x55: "SHORT_TERM_SECONDARY_OXYGEN_1_3",
    0x56: "LONG_TERM_SECONDARY_OXYGEN_1_3",
    0x57: "SHORT_TERM_SECONDARY_OXYGEN_2_4",
    0x58: "LONG_TERM_SECONDARY_OXYGEN_2_4",
    0x59: "FUEL_RAIL_ABS_PRESSURE",
    0x5A: "RELATIVE_ACCELERATOR",
    0x5B: "HYBRID_BATTERY_PACK",
    0x5C: "ENGINE_OIL_TEMP",
    0x5D: "FUEL_INJECTION_TIMING",
    0x5E: "ENGINE_FUEL_RATE",
    0x5F: "EMISSION_REQUIREMENTS",
    0x60: "PIDS_SUPPORTED_61_80",
    0x61: "DEMAND_ENGINE",
    0x62: "ACTUAL_ENGINE",
    0x63: "ENGINE_REF_TORQUE",
    0x64: "ENGINE_PERCENT_TORQUE",
    0x65: "AUXILIARY_INPUT",
    0x66: "MASS_AIR_FLOW_SENSOR",
    0x67: "ENGINE_COOLANT_TEMP",
    0x68: "INTAKE_AIR_TEMP_SENSOR",
    0x69: "COMMANDED_EGR_ERROR",
    0x6A: "COMMANDED_DIESEL_INTAKE",
    0x6B: "EXHAUST_GAS_RECIRCULATION",
    0x6C: "COMMANDED_THROTTLE_ACTUATOR",
    0x6D: "FUEL_PRESSURE_CONTROL_SYSTEM",
    0x6E: "INJECTION_PRESSURE_CONTROL_SYSTEM",
    0x6F: "TURBOCHARGER_COMPRESSOR",
    0x70: "BOOST_PRESSURE",
    0x71: "VARIABLE_GEOMETRY",
    0x72: "WASTEGATE_CONTROL",
    0x73: "EXHAUST_PRESSURE",
    0x74: "TURBOCHARGER_RPM",
    0x75: "TURBOCHARGER_TEMP",
    0x76: "TURBOCHARGER_TEMP",
    0x77: "CHARGER_AIR_COOLER_TEMP",
    0x78: "EXHAUST_GAS_TEMP",
    0x79: "EXHAUST_GAS_TEMP",
    0x7A: "DIESEL_PARTICULATE_FILTER",
    0x7B: "DIESEL_PARTICULATE_FILTER",
    0x7C: "DIESEL_PARTICULATE_FILTER_TEMP",
    0x7D: "NOX_CTRL_AREA",
    0x7E: "PM_NTE_CTLR_AREA",
    0x7F: "ENGINE_RUN_TIME",
    0x80: "PIDS_SUPPORTED_81_A0",
    0x81: "ENGINE_RUNTIME_AUXILIARY",
    0x82: "ENGINE_RUNTIME_AUXILIARY",
    0x83: "NOX_SENSOR",
    0x84: "MANIFOLD_SURFACE_TEMP",
    0x85: "NOX_REAGENT_SYSTEM",
    0x86: "PARTICULATE_MATTER_SENSOR",
    0x87: "INTAKE_MANIFOLD_PRESSURE",
    0xA0: "PIDS_SUPPORTED_A1_C0",
    0xC0: "PIDS_SUPPORTED_C1_E0",
    0xC3: "?",
    0xC4: "?"
}

OBD2_MODE_02_PIDS = {
    0x02: "DTC_FREEZE_STORED",
}

OBD2_MODE_05_PIDS = {
    0x0100: "OBD_MONITOR_IDS",
    0x0101: "O2_SENSOR_1_1",
    0x0102: "O2_SENSOR_1_2",
    0x0103: "O2_SENSOR_1_3",
    0x0104: "O2_SENSOR_1_4",
    0x0105: "O2_SENSOR_2_1",
    0x0106: "O2_SENSOR_2_2",
    0x0107: "O2_SENSOR_2_3",
    0x0108: "O2_SENSOR_2_4",
    0x0109: "O2_SENSOR_3_1",
    0x010A: "O2_SENSOR_3_2",
    0x010B: "O2_SENSOR_3_3",
    0x010C: "O2_SENSOR_3_4",
    0x010D: "O2_SENSOR_4_1",
    0x010E: "O2_SENSOR_4_2",
    0x010F: "O2_SENSOR_4_3",
    0x0110: "O2_SENSOR_4_4",
    0x0201: "O2_SENSOR_1_1",
    0x0202: "O2_SENSOR_1_2",
    0x0203: "O2_SENSOR_1_3",
    0x0204: "O2_SENSOR_1_4",
    0x0205: "O2_SENSOR_2_1",
    0x0206: "O2_SENSOR_2_2",
    0x0207: "O2_SENSOR_2_3",
    0x0208: "O2_SENSOR_2_4",
    0x0209: "O2_SENSOR_3_1",
    0x020A: "O2_SENSOR_3_2",
    0x020B: "O2_SENSOR_3_3",
    0x020C: "O2_SENSOR_3_4",
    0x020D: "O2_SENSOR_4_1",
    0x020E: "O2_SENSOR_4_2",
    0x020F: "O2_SENSOR_4_3",
    0x0210: "O2_SENSOR_4_4"
}

OBD2_MODE_0A_PIDS = {
    0x00: "MODE_9_SUPPORTED_PIDS_1_20",
    0x01: "VIN_MESSAGE_COUNT",
    0x02: "VIN",
    0x03: "CALIBRATION_ID_MESSAGE_COUNT",
    0x04: "CALLIBRATION_ID",
    0x05: "CVN_MESSAGE_COUNT",
    0x06: "CVN",
    0x07: "PERFORMANCE_TRACKING_MESSAGE_COUNT",
    0x08: "PERFORMANCE_TRACKING",
    0x09: "ECU_NAME_MESSAGE_COUNT",
    0x0A: "ECU_NAME",
    0x0B: "PERFORMANCE_TRACKING_COMPRESSION"
}

OBD2_MODE_DESCR = {
    0x01: "OBD2_MODE_SHOW_CURRENT_DATA1",
    0x02: "OBD2_MODE_SHOW_CURRENT_DATA2",
    0x03: "OBD2_MODE_SHOW_STORED_TROUBLECODES",
    0x04: "OBD2_MODE_CLEAR_DIAGNOSTIC_TROUBLECODES",
    0x05: "OBD2_MODE_TEST_RESULTS_OXYGEN",
    0x06: "OBD2_MODE_TEST_RESULTS_OTHER",
    0x07: "OBD2_MODE_SHOW_PENDING_DIAGNOSTIC_TROUBLECODES",
    0x08: "OBD2_MODE_CONTROL_OPERATION",
    0x09: "OBD2_MODE_REQUEST_VEHICLE_INFO",
    0x0A: "OBD2_MODE_PERMANENT_DIAGNOSTIC_TROUBLECODES"
}

for obdid, name in OBD2_MODE_DESCR.items():
    setattr(module_this, name, obdid)

# Diagnostic and Communications Management
UDS_SID_DESCR = {
    0x10: "UDS_SID_DIAGNOSTIC_SESSION_CONTORL",
    0x11: "UDS_SID_ECU_RESET",
    0x12: "UDS_SID_GMLAN_READ_FAILURE_RECORD",
    0x14: "UDS_SID_CLEAR_DIAGNOSTIC_INFORMATION",
    0x19: "UDS_SID_READ_DTC_INFORMATION",
    0x1A: "UDS_SID_GMLAN_READ_DIAGNOSTIC_ID",
    0x20: "UDS_SID_RETURN_TO_NORMAL",
    0x22: "UDS_SID_READ_DATA_BY_ID",
    0x23: "UDS_SID_READ_MEMORY_BY_ADDRESS",
    0x24: "UDS_SID_READ_SCALING_DATA_BY_ID",
    0x27: "UDS_SID_SECURITY_ACCESS",
    0x28: "UDS_SID_COMMUNICATION_CONTROL",
    0x2A: "UDS_SID_READ_DATA_BY_PERIODIC_ID",
    0x2C: "UDS_SID_DYNAMICALLY_DEFINE_DATA_ID",
    0x2D: "UDS_SID_DEFINE_PID_BY_MEMORY_ADDRESS",
    0x2E: "UDS_SID_WRITE_DATA_BY_ID",
    0x2F: "UDS_SID_IO_CONTORL_BY_ID",
    0x31: "UDS_SID_ROUTINE_CONTROL",
    0x34: "UDS_SID_REQUEST_DOWNLOAD",
    0x35: "UDS_SID_REQUEST_UPLOAD",
    0x36: "UDS_SID_TRANSFER_DATA",
    0x37: "UDS_SID_REQUEST_TANSFER_EXIT",
    0x38: "UDS_SID_REQUEST_FILE_TRANSFER",
    0x3B: "UDS_SID_GMLAN_WRITE_DID",
    0x3D: "UDS_SID_WRITE_MEMORY_BY_ADDRESS",
    0x3E: "UDS_SID_TESTER_PRESENT",
    0x83: "UDS_SID_ACCESS_TIMING_PARAMETER",
    0x84: "UDS_SID_SECURED_TRANSMISSION",
    0x85: "UDS_SID_CONTORL_DTC_SETTING",
    0x86: "UDS_SID_RESPONSE_ON_EVENT",
    0x87: "UDS_SID_LINK_CONTORL",
    0xA2: "UDS_SID_GMLAN_REPORT_PROGRAMMING_STATE",
    0xA5: "UDS_SID_GMLAN_ENTER_PROGRAMMING_MODE",
    0xA9: "UDS_SID_GMLAN_CHECK_CODES",
    0xAA: "UDS_SID_GMLAN_READ_DPID",
    0xAE: "UDS_SID_GMLAN_DEVICE_CONTROL",
}

for udsid, name in UDS_SID_DESCR.items():
    setattr(module_this, name, udsid)

# UDS = 0x7F [SID_requested] [NRC]
UDS_NRC_DESCR = {
    0x10: "UDS_NRC_GENERAL_REJECT",
    0x11: "UDS_NRC_SERVICE_NOT_SUPPORTED",
    0x12: "UDS_NRC_SUBFUNCTION_NOT_SUPPORTED",
    0x13: "UDS_NRC_INCORRECT_MESSAGE_LENGTH_OR_FORMAT",
    0x14: "UDS_NRC_RESPONSE_TOO_BIG",
    0x21: "UDS_NRC_BUSY_REPEAT_REQUEST",
    0x22: "UDS_NRC_CONDITION_NOT_CORRECT",
    0x24: "UDS_NRC_REQUEST_SEQUENCE_ERROR",
    0x25: "UDS_NRC_NONRESPONSE_FROM_SUBNET_COMPONENT",
    0x26: "UDS_NRC_FAILURE_PREVENTS_EXEC_OR_ACTION",
    0x31: "UDS_NRC_REQUEST_OUT_OF_RANGE",
    0x33: "UDS_NRC_SECURITY_ACCESS_DENIED",
    0x35: "UDS_NRC_INVALID_KEY",
    0x36: "UDS_NRC_EXCEEDED_NUMBER_OF_ATTEMPTS",
    0x37: "UDS_NRC_REQUIRED_TIME_DELEAY_NOT_EXPIRED",
    0x70: "UDS_NRC_UPLOAD_DOWNLOAD_NOT_ACCEPTED",
    0x71: "UDS_NRC_TRANSFER_DATA_SUSPENDED",
    0x72: "UDS_NRC_GENERAL_PROGRAMMING_FAILURE",
    0x73: "UDS_NRC_WRONG_BLOCK_SEQUENCE_COUNTER",
    0x78: "UDS_NRC_REQUEST_CORRECTLY_RESPONSE_PENDING",
    0x7E: "UDS_NRC_SUBFUNCTION_NOT_SUPPORTED_IN_SESSION",
    0x7F: "UDS_NRC_SERVICE_NOT_SUPPORTED_IN_SESSION"
}

for x in range(0x38, 0x4F):
    UDS_NRC_DESCR[x] = "UDS_NRC_RESERVED_%X" % x

for udsid, name in UDS_NRC_DESCR.items():
    setattr(module_this, name, udsid)


class OBD2(packetracer.Packet):
    __hdr__ = (
        ("mode", "B", 0),
        ("pid", "B", 0)
    )


class UDS(packetracer.Packet):
    __hdr__ = (
        ("sid", "B", 0),
        ("lev", "B", 0)
    )

    def _dissect(self, buf):
        # TODO: lev not always present
        return 2


ISOTP_TYPE_SF = 0x00  # single frame
ISOTP_TYPE_FF = 0x01  # first frame
ISOTP_TYPE_CF = 0x02  # consecutive frame
ISOTP_TYPE_FC = 0x03  # flow control

types_isotp_offset_upper = {
    ISOTP_TYPE_SF: 1,
    ISOTP_TYPE_FF: 2,
    ISOTP_TYPE_CF: 1,
    ISOTP_TYPE_FC: 3
}

types_isotp_offset_upper_got_type = {ISOTP_TYPE_SF, ISOTP_TYPE_FF}


class ISOTPBase(packetracer.Packet):
    def __get_sig(self):
        return (self.pci & 0xF0) >> 4

    def __set_sig(self, value):
        self.pci = (value & 0xF) << 4 | (self.pci & 0xF)

    sig = property(__get_sig, __set_sig)

    def _dissect(self, buf):
        # logger.debug("dissect in base class")
        # OBD/UDS can one be differentiated in ISOTP_TYPE_SF and ISOTP_TYPE_FF
        sig = buf[0] >> 4
        # logger.debug("bytes in ISOTP: %r" % buf)
        # logger.debug("ISOTP type: %X, offset: %d" % (sig, types_isotp_offset_upper[sig]))
        # logger.debug("upper bytes: %r" % buf[types_isotp_offset_upper[sig]: ])

        # check by request/response SID, on both OBD2 and UDS response will be SID+0x40
        if sig in types_isotp_offset_upper_got_type or (sig - 0x40) in types_isotp_offset_upper_got_type:
            obd_mode = buf[types_isotp_offset_upper[sig]]

            if obd_mode in OBD2_MODE_DESCR:
                # assume OBD2
                # logger.debug("got OBD2")
                self._init_handler(0, buf[types_isotp_offset_upper[sig]:])
            else:
                # assume UDS
                # logger.debug("got UDS, will use bytes: %r" % buf[types_isotp_offset_upper[sig]: ])
                self._init_handler(1, buf[types_isotp_offset_upper[sig]:])

        return types_isotp_offset_upper[sig]


class ISOTPSingleFrame(ISOTPBase):
    __hdr__ = (
        ("pci", "B", ISOTP_TYPE_SF),
    )

    def __get_dl(self):
        return self.pci & 0xF

    def __set_dl(self, value):
        self.pci = (self.pci & 0xF0) | (value & 0xF)

    dl = property(__get_dl, __set_dl)

    __handler__ = {
        0: OBD2, 1: UDS
    }


class ISOTPFirstFrame(ISOTPBase):
    __hdr__ = (
        ("pci", "H", ISOTP_TYPE_FF),
    )

    __handler__ = {
        0: OBD2, 1: UDS
    }

    def __get_sig(self):
        return (self.pci & 0xF000) >> 12

    def __set_sig(self, value):
        self.pci = (value & 0xF) << 12 | (self.pci & 0xFFF)

    sig = property(__get_sig, __set_sig)

    def __get_dl(self):
        return self.pci & 0xFFF

    def __set_dl(self, value):
        self.pci = (self.pci & 0xF000) | (value & 0xFFF)

    dl = property(__get_dl, __set_dl)


class ISOTPConsecutiveFrame(ISOTPBase):
    __hdr__ = (
        ("pci", "B", ISOTP_TYPE_CF),
    )

    __handler__ = {
        0: OBD2, 1: UDS
    }

    def __get_sn(self):
        return self.pci & 0xF

    def __set_sn(self, value):
        self.pci = (self.pci & 0xF0) | (value & 0xF)

    sn = property(__get_sn, __set_sn)


class ISOTPFlowControl(ISOTPBase):
    __hdr__ = (
        ("pci", "B", ISOTP_TYPE_FC),
        ("pci_blocksize", "B", 0),
        ("pci_minsep", "B", 0),
    )

    __handler__ = {
        0: OBD2, 1: UDS
    }

    def __get_flowstatus(self):
        return self.pci & 0xF

    def __set_flowstatus(self, value):
        self.pci = (self.pci & 0xF0) | (value & 0xF)

    flowstatus = property(__get_flowstatus, __set_flowstatus)


isotp_type_class = {
    ISOTP_TYPE_SF: ISOTPSingleFrame,
    ISOTP_TYPE_FF: ISOTPFirstFrame,
    ISOTP_TYPE_CF: ISOTPConsecutiveFrame,
    ISOTP_TYPE_FC: ISOTPFlowControl
}

# CAN flags, little endian..wtf?
CAN_MASK_EXT = 0x80000000
CAN_MASK_EXT_SHIFT = 31
CAN_MASK_RTR = 0x40000000
CAN_MASK_RTR_SHIFT = 30
CAN_MASK_ERR = 0x20000000
CAN_MASK_ERR_SHIFT = 29
# big endian
MASK_ID = 0x1FFFFFFF
MASK_FLAGS = ~MASK_ID

ISOTP_PKT_CLZES = {ISOTPConsecutiveFrame, ISOTPFirstFrame, ISOTPFlowControl, ISOTPSingleFrame}


class CAN(packetracer.Packet):
    """
        Format:

    The field containing the CAN ID and flags is in network byte order (big-endian).
    The bottom 29 bits contain the CAN ID of the frame. The remaining bits are:

    0x20000000 - set if the frame is an error message rather than a data frame.
    0x40000000 - set if the frame is a remote transmission request frame.
    0x80000000 - set if the frame is an extended 29-bit frame rather than a standard 11-bit frame. frame.
    Source: http://www.tcpdump.org/linktypes/LINKTYPE_CAN_SOCKETCAN.html


    SocketCan Packet, see https://www.kernel.org/doc/Documentation/networking/can.txt

    struct can_frame {
        canid_t can_id;  /* 32 bit CAN_ID + EFF/RTR/ERR flags */
        __u8    can_dlc; /* frame payload length in byte (0 .. 8) */
        __u8    __pad;   /* padding */
        __u8    __res0;  /* reserved / padding */
        __u8    __res1;  /* reserved / padding */
        __u8    data[8] __attribute__((aligned(8)));
    };
    special address description flags for the CAN_ID
        CAN_EFF_FLAG 0x80000000U /* EFF/SFF is set in the MSB */
        CAN_RTR_FLAG 0x40000000U /* remote transmission request */
        CAN_ERR_FLAG 0x20000000U /* error message frame */

    Native CAN structure (on wire):

    CAN lengths (11-bit ID):
    - Arbitration Field (12)
    -- ID: unique id/prio (11)
    -- Remote transmission request: 0/dominant for data, 1 for remote request (1)
    - Control (6)
    -- ID ext.: 0 for 11-bit ID (1)
    -- Reserved: must be 0 (1)
    -- Data length code: Number of bytes, 0 up to 8 (4)
    - Data: 0 up to 8 bytes (8)

    CAN lengths (29-bit ID):
    - Arbitration Field (40)
    -- ID: unique id/prio (11)
    -- Substitute remote request, must be 1 (1)
    -- Identifier extension bit: must be 1 (1)
    -- ID: unique id/prio (18)
    -- Remote transmission request: 0/dominant for data, 1 for remote request (1)
    - Control (6)
    -- Reserved: must be 0 (2)
    -- Data length code: Number of bytes, 0 up to 8 (4)
    - Data: 0 up to 8 bytes (8)

    """
    __hdr__ = (
        ("flag_id", "I", 0),
        ("dlc", "B", 0),
        ("pad", "B", 0),
        ("res1", "B", 0),
        ("res2", "B", 0)
    )

    __handler__ = {
        ISOTP_TYPE_SF: ISOTPSingleFrame,
        ISOTP_TYPE_FF: ISOTPFirstFrame,
        ISOTP_TYPE_CF: ISOTPConsecutiveFrame,
        ISOTP_TYPE_FC: ISOTPFlowControl
    }

    def __get_extended(self):
        return 0 if (self.flag_id & CAN_MASK_EXT) == 0 else 1

    def __set_extended(self, value):
        self.flag_id = (self.flag_id & ~CAN_MASK_EXT) | ((value & 1) << CAN_MASK_EXT_SHIFT)

    extended = property(__get_extended, __set_extended)

    def __get_rtr(self):
        return 0 if (self.flag_id & CAN_MASK_RTR) == 0 else 1

    def __set_rtr(self, value):
        self.flag_id = (self.flag_id & ~CAN_MASK_RTR) | ((value & 1) << CAN_MASK_RTR_SHIFT)

    rtr = property(__get_rtr, __set_rtr)

    def __get_err(self):
        return 0 if (self.flag_id & CAN_MASK_ERR) == 0 else 1

    def __set_err(self, value):
        self.flag_id = (self.flag_id & ~CAN_MASK_ERR) | ((value & 1) << CAN_MASK_ERR_SHIFT)

    err = property(__get_err, __set_err)

    def __get_id(self):
        return self.flag_id & MASK_ID

    def __set_id(self, value):
        flags_be = self.flag_id & MASK_FLAGS
        self.flag_id = flags_be | (value & MASK_ID)

    id = property(__get_id, __set_id)

    def _dissect(self, buf):
        # assume ISO-TP
        isotp_type = (buf[8] & 0xF0) >> 4
        # logger.debug("got ISOTP type: %d, class will be: %r" %
        # (isotp_type, isotp_type_class[isotp_type]))
        self._init_handler(isotp_type, buf[8:])
        return 8

    def _update_fields(self):
        if not self._header_changed:
            return

        if self.id > 0x7FF:
            self.extended = 1
        else:
            self.extended = 0

    def show_packet(self, prefix=""):
        if self.body_handler.__class__ not in ISOTP_PKT_CLZES:
            logger.warning("not an isotp packet: %r", self.body_handler)
            return

        isotp_pkt = self.body_handler
        output = []
        output.append("%s [ID: 0x%05X] " % (prefix, self.id))

        uds_pkt = self[UDS]

        if uds_pkt is not None:
            if uds_pkt.sid in UDS_SID_DESCR:
                output.append("-> %s" % UDS_SID_DESCR.get(uds_pkt.sid))
            elif uds_pkt.sid - 0x40 in UDS_SID_DESCR:
                output.append("<- %s" % UDS_SID_DESCR.get(uds_pkt.sid - 0x40))
            elif uds_pkt.sid == module_this.UDS_NRC_SERVICE_NOT_SUPPORTED_IN_SESSION:
                output.append("Err: %s" % UDS_NRC_DESCR.get(uds_pkt.bin()[2], "0x%X" % uds_pkt.bin()[2]))
            else:
                output.append("<unknown: %s>" % uds_pkt.sid)
            output.append(", content: %r" % uds_pkt.bin())
        else:
            output.append(" %r, content: %s" % (isotp_pkt.__class__.__name__, isotp_pkt.body_bytes))

        print("%s" % "".join(output))



"""Real-Time Transport Protocol"""

from packetracer import packetracer

# version 1100 0000 0000 0000 ! 0xC000	14
# p		  0010 0000 0000 0000 ! 0x2000	13
# x		  0001 0000 0000 0000 ! 0x1000	12
# cc	  0000 1111 0000 0000 ! 0x0F00	 8
# m		  0000 0000 1000 0000 ! 0x0080	 7
# pt	  0000 0000 0111 1111 ! 0x007F	 0
#

_VERSION_MASK = 0xC000
_P_MASK = 0x2000
_X_MASK = 0x1000
_CC_MASK = 0x0F00
_M_MASK = 0x0080
_PT_MASK = 0x007F
_VERSION_SHIFT = 14
_P_SHIFT = 13
_X_SHIFT = 12
_CC_SHIFT = 8
_M_SHIFT = 7
_PT_SHIFT = 0

VERSION = 2


class RTP(packetracer.Packet):
    __hdr__ = (
        ("type", "H", 0x8000),
        ("seq", "H", 0),
        ("ts", "I", 0),
        ("ssrc", "I", 0)
    )

    def getversion(self):
        return (self.type & _VERSION_MASK) >> _VERSION_SHIFT

    def setversion(self, value):
        self.type = (value << _VERSION_SHIFT) | (self.type & ~_VERSION_MASK)

    version = property(getversion, setversion)

    def getp(self):
        return (self.type & _P_MASK) >> _P_SHIFT

    def setp(self, value):
        self.type = (value << _P_SHIFT) | (self.type & ~_P_MASK)

    p = property(getp, setp)

    def getx(self):
        return (self.type & _X_MASK) >> _X_SHIFT

    def setx(self, value):
        self.type = (value << _X_SHIFT) | (self.type & ~_X_MASK)

    x = property(getx, setx)

    def getcc(self):
        return (self.type & _CC_MASK) >> _CC_SHIFT

    def setcc(self, value):
        self.type = (value << _CC_SHIFT) | (self.type & ~_CC_MASK)

    cc = property(getcc, setcc)

    def getm(self):
        return (self.type & _M_MASK) >> _M_SHIFT

    def setm(self, value):
        self.type = (value << _M_SHIFT) | (self.type & ~_M_MASK)

    m = property(getm, setm)

    def getpt(self):
        return (self.type & _PT_MASK) >> _PT_SHIFT

    def setpt(self, value):
        self.type = (value << _PT_SHIFT) | (self.type & ~_PT_MASK)

    pt = property(getpt, setpt)



"""Portmap / rpcbind."""

from packetracer import packetracer

PMAP_PROG = 100000
PMAP_PROCDUMP = 4
PMAP_VERS = 2


class Pmap(packetracer.Packet):
    __hdr__ = (
        ("prog", "I", 0),
        ("vers", "I", 0),
        ("prot", "I", 0),
        ("port", "I", 0),
    )



"""Address Resolution Protocol."""

from packetracer import packetracer

# Hardware address format
ARP_HRD_ETH = 0x0001  # ethernet hardware
ARP_HRD_IEEE802 = 0x0006  # IEEE 802 hardware

# Protocol address format
ARP_PRO_IP = 0x0800  # IP protocol

# ARP operation
ARP_OP_REQUEST = 1  # request to resolve ha given pa
ARP_OP_REPLY = 2  # response giving hardware address
ARP_OP_REVREQUEST = 3  # request to resolve pa given ha
ARP_OP_REVREPLY = 4  # response giving protocol address


class ARP(packetracer.Packet):
    __hdr__ = (
        ("hrd", "H", ARP_HRD_ETH),
        ("pro", "H", ARP_PRO_IP),
        ("hln", "B", 6),  # hardware address length
        ("pln", "B", 4),  # protocol address length
        ("op", "H", ARP_OP_REQUEST),
        ("sha", "6s", b"\x00" * 6),  # sender mac
        ("spa", "4s", b"\x00" * 4),  # sender ip
        ("tha", "6s", b"\x00" * 6),  # target mac
        ("tpa", "4s", b"\x00" * 4)  # target ip
    )

    # convenient access
    sha_s = packetracer.get_property_mac("sha")
    spa_s = packetracer.get_property_ip4("spa")
    tha_s = packetracer.get_property_mac("tha")
    tpa_s = packetracer.get_property_ip4("tpa")



"""Cisco Hot Standby Router Protocol."""

from packetracer import packetracer

# Opcodes
HELLO = 0
COUP = 1
RESIGN = 2

# States
INITIAL = 0x00
LEARN = 0x01
LISTEN = 0x02
SPEAK = 0x04
STANDBY = 0x08
ACTIVE = 0x10


class HSRP(packetracer.Packet):
    __hdr__ = (
        ("version", "B", 0),
        ("opcode", "B", 0),
        ("state", "B", 0),
        ("hello", "B", 0),
        ("hold", "B", 0),
        ("priority", "B", 0),
        ("group", "B", 0),
        ("rsvd", "B", 0),
        ("auth", "8s", b"cisco"),
        ("vip", "4s", b"")
    )



"""TriggerList for handling dynamic headers."""
import logging

logger = logging.getLogger("packetracer")

TRIGGERLIST_CONTENT_SIMPLE = {bytes, tuple}


class TriggerList(list):
    """
    List with trigger-capabilities representing a Packet header.
    This list can contain any type of raw bytes, tuples like (key, value)
    or packets. Calling bin() will reassemble a content like
    [b"somebytes", mypacket, ("tuplekey", "tuplevalue")]
    to this: b"somebytes" + mypacket.bin() + ("tuplekey", "tuplevalue")[1].
    Custom reassemblation for tuples can be done by overwriting "_pack()".
    """

    def __init__(self, packet, dissect_callback=None, buffer=b"", headerfield_name=""):
        """
        packet -- packet where this TriggerList gets integrated
        dissect_callback -- callback which dessects byte string b"buffer", returns [a, b, c, ...]
        buffer -- byte string to be dissected
        headerfield_name -- name of this triggerlist when placed in a packet
        """
        super().__init__()
        # set by external Packet
        # logger.debug(">>> init of TriggerList (contained in %s): %s" %
        # (packet.__class__.__name__, buffer))
        self._packet = packet
        self._dissect_callback = dissect_callback
        self._cached_result = buffer
        self._headerfield_name = headerfield_name

    def _lazy_dissect(self):
        if self._packet._unpacked == False:
            # Before changing TriggerList we need to unpack or
            # cached header won't fit on _unpack(...)
            # This is called before any changes to TriggerList so place it here.
            # Ignore if TriggerList changed in _dissect (_unpacked is None)
            self._packet._unpack()

        if self._dissect_callback is None:
            # already dissected, ignore
            return

        try:
            # logger.debug("Dissecting in TL")
            initial_list_content = self._dissect_callback(self._cached_result)
        except:
            # If anything goes wrong: raw bytes will be accessible in any case
            # logger.debug("Failed to dissect in TL")
            initial_list_content = [self._cached_result]

        self._dissect_callback = None
        # This is re-calling _lazy_dissect() but directly returning after second if
        # extend() is clearing cache -> remember cache
        cache_tmp = self._cached_result
        self.extend(initial_list_content)
        self._cached_result = cache_tmp

    # Python predefined overwritten methods

    def __getitem__(self, pos):
        self._lazy_dissect()
        return super().__getitem__(pos)

    def __iadd__(self, v):
        """Item can be added using '+=', use 'append()' instead."""
        self._lazy_dissect()
        super().__iadd__(v)
        self.__refresh_listener([v])
        return self

    def __setitem__(self, k, v):
        self._lazy_dissect()
        try:
            # Remove listener from old packet which gets overwritten
            self[k].remove_change_listener(None, remove_all=True)
        except:
            pass
        super().__setitem__(k, v)
        self.__refresh_listener([v])

    def __delitem__(self, k):
        # logger.debug("removing elements: %r" % k)
        self._lazy_dissect()
        if type(k) is int:
            itemlist = [self[k]]
        else:
            # Assume slice: [x:y]
            itemlist = self[k]
        super().__delitem__(k)
        # logger.debug("removed, handle mod")
        self.__refresh_listener(itemlist, connect_packet=False)

    # logger.debug("finished removing")

    def __len__(self):
        self._lazy_dissect()
        return super().__len__()

    def __iter__(self):
        self._lazy_dissect()
        return super().__iter__()

    def append(self, v):
        self._lazy_dissect()
        super().append(v)
        # logger.debug("handling mod")
        self.__refresh_listener([v])

    # logger.debug("finished")

    def extend(self, v):
        self._lazy_dissect()
        super().extend(v)
        self.__refresh_listener(v)

    def insert(self, pos, v):
        self._lazy_dissect()
        super().insert(pos, v)
        self.__refresh_listener([v])

    def clear(self):
        self._lazy_dissect()
        items = [item for item in self]
        super().clear()
        self.__refresh_listener(items, connect_packet=False)

    def __refresh_listener(self, val, connect_packet=True):
        """
        Handle modifications of this TriggerList (adding, removing, ...).
        WARNING: packets can only be put in one tl once at a time

        val -- list of bytes, tuples or packets
        connect_packet -- connect packet to this tl and parent packet
        """
        for v in val:
            # Ignore non-packets
            if type(v) in TRIGGERLIST_CONTENT_SIMPLE:
                continue

            if connect_packet:
                # Allow packet in TL to access packet containing this TL:
                # packet1( TL[packet2->"access to packet1"] )
                v._triggelistpacket_parent = self._packet
                # TriggerList observes changes on packets:
                # base packet <- TriggerList (observes changes, set changed status
                # in basepacket) <- contained packet (changes)
                # Add change listener to the packet this TL is contained in.
                lwrapper = lambda: self._notify_change()
                v._add_change_listener(lwrapper)
            else:
                # Remove any old listener
                v._remove_change_listener()
                # Remove old parent
                v._triggelistpacket_parent = None
        # logger.debug("refreshed listener!")
        self._notify_change()

    def _notify_change(self):
        """
        Update _header_changed and _header_format_changed of the Packet having
        this TriggerList as field and _cached_result.
        Called by: this list on changes or Packets in this list
        """
        # logger.debug("!!! Packet notified about update: %r -> %r" % (self._packet.__class__, self))

        self._packet._header_changed = True
        self._packet._header_format_changed = True
        # List changed: old cache of TriggerList not usable anymore
        self._cached_result = None

    def bin(self):
        """
        Output the TriggerLists elements as concatenated bytestring.
        Custom implementations for tuple-handling can be set by overwriting _pack().
        """
        # logger.debug("packing triggerlist content")
        # logger.debug("sep in TriggerList: %r" % self._packet.sep)

        if self._cached_result is None:
            result_arr = []
            entry_type = None

            for entry in self:
                entry_type = type(entry)
                # logger.debug("type is: %r" % entry_type)

                if entry_type is bytes:
                    result_arr.append(entry)
                elif entry_type is tuple:
                    result_arr.append(self._pack(entry))
                else:
                    # This Must be a packet, otherthise invalid entry!
                    result_arr.append(entry.bin())

            self._cached_result = b"".join(result_arr)
        # logger.debug("new cached result: %s" % self._cached_result)

        return self._cached_result

    def find_pos(self, search_cb, offset=0):
        """
        Find an item-position giving search callback as search criteria.

        search_cb -- callback to compare values, signature: callback(value) [True|False]
            Return True to return value found.
        offset -- start at index "offset" to search
        return -- index of first element found or None
        """
        self._lazy_dissect()
        while offset < len(self):
            try:
                if search_cb(self[offset]):
                    return offset
            except:
                # error on callback (unknown fields etc), ignore
                pass
            offset += 1
        # logger.debug("position not found")
        return None

    def find_value(self, search_cb, offset=0):
        """
        Same as find_pos() but directly returning found value or None.
        """
        self._lazy_dissect()
        try:
            return self[self.find_pos(search_cb, offset=offset)]
        except:
            return None

    def _pack(self, tuple_entry):
        """
        This can  be overwritten to convert tuples in TriggerLists to bytes (see layer567/http.py)
        return -- byte string representation of this tuple entry
            eg (b"Host", b"localhost") -> b"Host: localhost"
        """
        return tuple_entry[1]

    def __repr__(self):
        self._lazy_dissect()
        return super().__repr__()

    def __str__(self):
        self._lazy_dissect()
        tl_descr_l = []
        contains_pkt = False

        for val_tl in self:
            if type(val_tl) in TRIGGERLIST_CONTENT_SIMPLE:
                tl_descr_l.append("%s" % str(val_tl))
            else:
                # assume packet
                # pkt_fqn = val_tl.__module__[9:] + "." + val_tl.__class__.__name__
                # tl_descr_l.append(pkt_fqn)
                tl_descr_l.append("%s" % val_tl)
                contains_pkt = True

        if not contains_pkt or len(tl_descr_l) == 0:
            # Oneline output
            return "[" + ", ".join(tl_descr_l) + "]"
        else:
            # Multiline output
            final_descr = ["(see below)\n" + "-" * 10 + "\n"]

            for idx, val in enumerate(tl_descr_l):
                idx_descr = "%s[%d]" % (self._headerfield_name[1:], idx)
                final_descr.append("-> %s:\n%s\n" % (idx_descr, val))
            final_descr.append("-" * 10)
            return "".join(final_descr)

    def get_by_key(self, key_needle, idx_startat=0):
        """
        Allow retrieving the value of a tuple key/value-pair.
        This isn't done via dictionaries as keys don't have to be unique.
        This isn't done by __getitem__ either because it would be to ambiguous
        in contrast to index access.

        key -- The key to ssearch for a value like (key, value). The search
            is case INsensitive!
        return -- First matching value corresponding to key like idx, b"value"
            or None, None if nothing was found

        """
        try:
            key_needle = key_needle.lower()
        except:
            # not a string
            pass

        idx = self.find_pos(
            search_cb=lambda tpl: tpl[0] == key_needle or tpl[0].lower() == key_needle,
            offset=idx_startat)

        if idx is None:
            return None, None
        return idx, self[idx][1]

    def set_by_key(self, key_needle, value, idx_startat=0):
        """
        Do inverse of get_by_key()
        """
        try:
            key_needle = key_needle.lower()
        except:
            # not a string
            pass

        idx = self.find_pos(
            search_cb=lambda tpl: tpl[0] == key_needle or tpl[0].lower() == key_needle,
            offset=idx_startat)

        if idx is None:
            return
        # Update using original key
        self[idx] = (self[idx][0], value)



"""Dynamic Trunking Protocol."""
from packetracer import packetracer, triggerlist
from packetracer.structcbs import unpack_HH

TRUNK_NAME = 0x01
MAC_ADDR = 0x04


class DTP(packetracer.Packet):
    __hdr__ = (
        ("v", "B", 0),
        ("tvs", None, triggerlist.TriggerList)
    )

    def _dissect(self, buf):
        off = 1
        dlen = len(buf)
        tvs = []

        while off < dlen:
            # length: inclusive header
            _, hlen = unpack_HH(buf[off: off + 4])
            packet = TV(buf[off: off + hlen])
            tvs.append(packet)
            off += hlen

        self.tvs.extend(tvs)
        return off


class TV(packetracer.Packet):
    __hdr__ = (
        ("t", "H", 0),
        ("len", "H", 0)
    )
